"""
Import ConceptNet file to KGTK.
"""


import sys

def parser():
    return {
        'help': 'Import ConceptNet into KGTK.' 
    }


def add_arguments(parser):
    """
    Parse arguments
    Args:
            parser (argparse.ArgumentParser)
    """
    # '$label == "/r/DefinedAs" && $node2=="/c/en/number_zero"'
    parser.add_argument(action="store", type=str, dest="filename", metavar='filename', help='filename here')
    parser.add_argument('--english_only', action="store_true", help="Only english conceptnet?")
    parser.add_argument('--sort', action="store_true", help="Should we sort the file on s-p-o?")


def run(filename, english_only, sort):
    # import modules locally
    import sys # type: ignore
    from kgtk.exceptions import kgtk_exception_auto_handler
    import csv

    def row_to_edge(row):
        return '\t'.join(row) + '\n'

    try:
        in_columns=['assertion','rel','subj','obj','metadata']
        out_columns=['node1', 'label', 'node2']
        with open(filename, 'r') as f:
            reader = csv.reader(f, delimiter='\t', quotechar='"')
            sys.stdout.write(row_to_edge(out_columns))
            for row in reader:
                new_row=[row[2], row[1], row[3]]
                if not english_only or (new_row[0].startswith('/c/en/') and new_row[2].startswith('/c/en/')):
                    sys.stdout.write(row_to_edge(new_row))

    except Exception as e:
            kgtk_exception_auto_handler(e)



import sys
import re
from typing import TextIO
from kgtk.exceptions import KGTKException
from etk.wikidata.entity import WDItem, WDProperty
from etk.etk_module import ETKModule
from etk.etk import ETK
from etk.knowledge_graph import KGSchema
from etk.wikidata import wiki_namespaces
from etk.wikidata.statement import Rank
import rfc3986
import json
from etk.wikidata.value import (
    Precision,
    Item,
    StringValue,
    TimeValue,
    QuantityValue,
    MonolingualText,
    GlobeCoordinate,
    ExternalIdentifier,
    URLValue
)
from etk.knowledge_graph.node import LiteralType

BAD_CHARS = [":", "&", ",", " ",
             "(", ")", "\'", '\"', "/", "\\", "[", "]", ";", "|"]

class Generator:
    def __init__(self,**kwargs):
        label_set = kwargs.pop("label_set")
        description_set = kwargs.pop("description_set")
        alias_set = kwargs.pop("alias_set")
        n = int(kwargs.pop("n"))
        warning = kwargs.pop("warning")
        log_path = kwargs.pop("log_path")
        # set sets
        self.set_sets(label_set,description_set,alias_set)
        # column name order_map
        self.order_map = {}
        self.n = n
        self.yyyy_mm_dd_pattern = re.compile(
            r"[12]\d{3}-(0[1-9]|1[0-2])-(0[1-9]|[12]\d|3[01])")
        self.yyyy_pattern = re.compile(r"[12]\d{3}")
        self.quantity_pattern = re.compile(
            r"([\+|\-]?[0-9]+\.?[0-9]*[e|E]?[\-]?[0-9]*)(?:\[([\+|\-]?[0-9]+\.?[0-9]*),([\+|\-]?[0-9]+\.?[0-9]*)\])?([U|Q](?:.*))?")
        self.warning = warning
        if self.warning:
            self.warn_log = open(log_path,"w")
        self.to_append_statement_id = None
        self.corrupted_statement_id = None
        self.to_append_statement = None # for Json generator
    def serialize(self):
        raise NotImplemented
    def finalize(self):
        if self.warning:
            self.warn_log.close()
        self.serialize()
    def set_sets(self,label_set:str,description_set:str,alias_set:str):
        self.label_set, self.alias_set, self.description_set = set(label_set.split(",")), set(alias_set.split(",")), set(description_set.split(","))
    
    def initialize_order_map(self, edge:str):
        edge_list = edge.strip("\n").split("\t")
        node1_index = edge_list.index("node1")
        node2_index = edge_list.index("node2")
        prop_index = edge_list.index("label")
        id_index = edge_list.index("id")
        if not all([node1_index > -1, node2_index > -1, prop_index > -1, id_index > -1]):
            raise KGTKException(
                "Header of kgtk file misses at least one of required column names: (node1, node2, property and id)")
        else:
            self.order_map["node1"] = node1_index
            self.order_map["node2"] = node2_index
            self.order_map["label"] = prop_index
            self.order_map["id"] = id_index

    @staticmethod
    def process_text_string(string: str) -> [str, str]:
        ''' 
        Language detection is removed from triple generation. The user is responsible for detect the language
        '''
        if len(string) == 0:
            return ["", "en"]
        if "@" in string:
            res = string.split("@")
            text_string = "@".join(res[:-1]).replace('"', "").replace("'", "")
            lang = res[-1].replace('"', '').replace("'", "")
            if len(lang) > 2:
                lang = "en"
        else:
            text_string = string.replace('"', "").replace("'", "")
            lang = "en"
        return [text_string, lang]
    
    @staticmethod
    def is_invalid_decimal_string(num_string)->bool:
        '''
        if a decimal string too small, return True TODO
        '''
        if num_string == None:
            return False
        else:
            if abs(float(num_string)) < 0.0001 and float(num_string) != 0:
                return True
            return False

    @staticmethod
    def is_valid_uri_with_scheme_and_host(uri: str)->bool:
        '''
        https://github.com/python-hyper/rfc3986/issues/30#issuecomment-461661883
        '''
        try:
            uri = rfc3986.URIReference.from_string(uri)
            rfc3986.validators.Validator().require_presence_of(
                "scheme", "host").check_validity_of("scheme", "host").validate(uri)
            return True
        except:
            return False

    @staticmethod
    def clean_number_string(num:str)->str:
        from numpy import format_float_positional
        if num == None:
            return None
        else:
            return format_float_positional(float(num), trim="-")
    
    @staticmethod
    def replace_illegal_string(s: str) -> str:
        '''
        this function serves as the last gate of keeping illegal characters outside of entity creation.
        '''
        for char in BAD_CHARS:
            s = s.replace(char, "_")
        return s

class TripleGenerator(Generator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        prop_file = kwargs.pop("prop_file")
        prop_declaration = kwargs.pop("prop_declaration")
        dest_fp = kwargs.pop("dest_fp")
        truthy = kwargs.pop("truthy")
        use_id = kwargs.pop("use_id")
        self.datatype_mapping = {
            "item": Item,
            "time": TimeValue,
            "globe-coordinate": GlobeCoordinate,
            "quantity": QuantityValue,
            "monolingualtext": MonolingualText,
            "string": StringValue,
            "external-identifier": ExternalIdentifier,
            "url": StringValue,
            "property":WDProperty,
        }
        self.prop_declaration = prop_declaration
        self.set_properties(prop_file)
        self.fp = dest_fp
        self.read_num_of_lines = 0
        self.truthy = truthy
        self.reset_etk_doc()
        self.serialize_prefix()
        self.use_id = use_id

    
    def parse_edges(self,edge:str):
        # use the order_map to map the node
        edge_list = edge.strip("\n").split("\t")
        node1 = edge_list[self.order_map["node1"]].strip()
        node2 = edge_list[self.order_map["node2"]].strip()
        prop = edge_list[self.order_map["label"]].strip()
        e_id = edge_list[self.order_map["id"]].strip()  
        return node1, node2, prop, e_id
    
    def read_prop_declaration(self,line_number:int, edge:str):
        node1, node2, prop, e_id = self.parse_edges(edge)
        if prop == "data_type":
            self.prop_types[node1] = self.datatype_mapping[node2.strip()]
        return
    
    def set_properties(self,prop_file:str):
        self.prop_types = {}
        if prop_file == "NONE":
            return
        
        with open(prop_file, "r") as fp:
            props = fp.readlines()
        for line in props[1:]:
            node1, _, node2 = line.split("\t")
            try:
                self.prop_types[node1] = self.datatype_mapping[node2.strip()]
            except:
                raise KGTKException(
                    "DataType {} of node {} is not supported.\n".format(
                        node2, node1
                    )
                )
    
    def _node_2_entity(self, node: str):
        '''
        A node can be Qxxx or Pxxx, return the proper entity.
        '''
        if node in self.prop_types:
            entity = WDProperty(node, self.prop_types[node])
        else:
            entity = WDItem(TripleGenerator.replace_illegal_string(node))
        return entity
    
    def reset_etk_doc(self, doc_id: str = "http://isi.edu/default-ns/projects"):
        """
        reset the doc object and return it. Called at initialization and after outputting triples.
        """
        kg_schema = KGSchema()
        kg_schema.add_schema("@prefix : <http://isi.edu/> .", "ttl")
        self.etk = ETK(kg_schema=kg_schema, modules=ETKModule)
        self.doc = self.etk.create_document({}, doc_id=doc_id)
        # TODO support customized namespace binding
        for k, v in wiki_namespaces.items():
            self.doc.kg.bind(k, v)

    def serialize(self):
        """
        Seriealize the triples. Used a hack to avoid serializing the prefix again.
        """
        docs = self.etk.process_ems(self.doc)
        self.fp.write("\n\n".join(
            docs[0].kg.serialize("ttl").split("\n\n")[1:]))
        self.fp.flush()
        self.reset()
    
    def serialize_prefix(self):
        """
        This function should be called only once after the doc object is initialized.
        In order to serialize the prefix at the very begining it has to be printed per the change of rdflib 4.2.2->5.0.0
        Relevent issue: https://github.com/RDFLib/rdflib/issues/965
        """
        for k, v in wiki_namespaces.items():
            line = "@prefix " + k + ": <" + v + "> .\n"
            self.fp.write(line)
        self.fp.write("\n")
        self.fp.flush()
        self.reset()

    def reset(self):
        self.to_append_statement_id = None
        self.to_append_statement = None
        self.read_num_of_lines = 0
        self.reset_etk_doc()
    
    def generate_label_triple(self, node1: str, node2: str) -> bool:
        entity = self._node_2_entity(node1)
        text_string, lang = TripleGenerator.process_text_string(node2)
        entity.add_label(text_string, lang=lang)
        self.doc.kg.add_subject(entity)
        return True

    def generate_description_triple(self, node1: str, node2: str) -> bool:
        entity = self._node_2_entity(node1)
        text_string, lang = TripleGenerator.process_text_string(node2)
        entity.add_description(text_string, lang=lang)
        self.doc.kg.add_subject(entity)
        return True

    def generate_alias_triple(self, node1: str, node2: str) -> bool:
        entity = self._node_2_entity(node1)
        text_string, lang = TripleGenerator.process_text_string(node2)
        entity.add_alias(text_string, lang=lang)
        self.doc.kg.add_subject(entity)
        return True

    def generate_prop_declaration_triple(self, node1: str, node2: str) -> bool:
        # update the known prop_types
        if node1 in self.prop_types:
            if not self.prop_declaration:
                raise KGTKException("Duplicated property definition of {} found!".format(node1))
        else:
            self.prop_types[node1] = node2
        
        prop = WDProperty(node1, self.datatype_mapping[node2])
        self.doc.kg.add_subject(prop)
        return True

    def generate_normal_triple(
            self, node1: str, property: str, node2: str, is_qualifier_edge: bool, e_id: str) -> bool:
        if self.use_id:
            e_id = TripleGenerator.replace_illegal_string(e_id)
        entity = self._node_2_entity(node1)
        edge_type = self.prop_types[property]
        if edge_type == Item:
            object = WDItem(TripleGenerator.replace_illegal_string(node2))
        elif edge_type == WDProperty:
            object = WDProperty(TripleGenerator.replace_illegal_string(node2),self.prop_types[node2])
        
        elif edge_type == TimeValue:
            if self.yyyy_mm_dd_pattern.match(node2):
                try:
                    dateTimeString = node2
                    object = TimeValue(
                        value=dateTimeString,  # TODO
                        calendar=Item("Q1985727"),
                        precision=Precision.year,
                        time_zone=0,
                    )
                except:
                    return False
            elif self.yyyy_pattern.match(node2):
                try:
                    dateTimeString = node2 + "-01-01"
                    object = TimeValue(
                        value=dateTimeString,  # TODO
                        calendar=Item("Q1985727"),
                        precision=Precision.year,
                        time_zone=0,
                    )
                except:
                    return False
            else:
                try:
                    # TODO, in future, the two cases above will be dropped in principle to comply with the iso format
                    # now it is iso format
                    assert(node2[0] == "^")
                    node2 = node2[1:] # remove ^
                    if node2.startswith("+"):
                        node2 = node2[1:]
                    dateTimeString, precision = node2.split("/")
                    dateTimeString = dateTimeString[:-1] # remove Z
                    object = TimeValue(
                        value=dateTimeString,
                        calendar=Item("Q1985727"),
                        precision=precision,
                        time_zone=0,
                    )
                except:
                    return False

        elif edge_type == GlobeCoordinate:
            latitude, longitude = node2[1:].split("/")
            latitude = float(latitude)
            longitude = float(longitude)
            object = GlobeCoordinate(
                latitude, longitude, 0.0001, globe=Item("Q2")) # earth

        elif edge_type == QuantityValue:
            # +70[+60,+80]Q743895
            res = self.quantity_pattern.match(node2).groups()
            amount, lower_bound, upper_bound, unit = res

            amount = TripleGenerator.clean_number_string(amount)
            num_type = self.xsd_number_type(amount)
            
            lower_bound = TripleGenerator.clean_number_string(lower_bound)
            upper_bound = TripleGenerator.clean_number_string(upper_bound)
            if unit != None:
                if upper_bound != None and lower_bound != None:
                    object = QuantityValue(amount, unit=Item(
                        unit), upper_bound=upper_bound, lower_bound=lower_bound, type=num_type)
                else:
                    object = QuantityValue(amount, unit=Item(unit), type=num_type)
            else:
                if upper_bound != None and lower_bound != None:
                    object = QuantityValue(
                        amount, upper_bound=upper_bound, lower_bound=lower_bound, type=num_type)
                else:
                    object = QuantityValue(amount, type=num_type)

        elif edge_type == MonolingualText:
            text_string, lang = TripleGenerator.process_text_string(node2)
            object = MonolingualText(text_string, lang)
        elif edge_type == ExternalIdentifier:
            object = ExternalIdentifier(node2)
        elif edge_type == URLValue:
            if TripleGenerator.is_valid_uri_with_scheme_and_host(node2):
                object = URLValue(node2)
            else:
                return False
        else:
            # treat everything else as stringValue
            object = StringValue(node2)

        if type(object) == WDItem or type(object) == WDProperty:
            self.doc.kg.add_subject(object)

        if is_qualifier_edge:
            # edge: e8 p9 ^2013-01-01T00:00:00Z/11
            # create qualifier edge on previous STATEMENT and return the updated STATEMENT
            self.to_append_statement.add_qualifier(property, object)
            self.doc.kg.add_subject(self.to_append_statement)
        else:
            # edge: q1 p8 q2 e8
            # create brand new property edge and replace STATEMENT
            if self.truthy:
                self.to_append_statement = entity.add_truthy_statement(
                    property, object, statement_id=e_id) if self.use_id else entity.add_truthy_statement(property, object)
            else:
                self.to_append_statement = entity.add_statement(
                    property, object, statement_id=e_id) if self.use_id else entity.add_statement(property, object)
            self.doc.kg.add_subject(entity)
        return True
    

    def entry_point(self, line_number: int, edge: str):
        """
        generates a list of two, the first element is the determination of the edge type using corresponding edge type
        the second element is a bool indicating whether this is a valid property edge or qualifier edge.
        Call corresponding downstream functions
        """
        if line_number == 1:
            # initialize the order_map
            self.initialize_order_map(edge)
            return

        # use the order_map to map the node
        node1, node2, prop, e_id = self.parse_edges(edge)
        if line_number == 2:
            # by default a statement edge
            is_qualifier_edge = False
        else:
            if node1 != self.to_append_statement_id and node1 != self.corrupted_statement_id:
                is_qualifier_edge = False
                # also a new statement edge
                if self.read_num_of_lines >= self.n:
                    self.serialize()
            else:
                # qualifier edge or property declaration edge
                is_qualifier_edge = True
                if node1 == self.corrupted_statement_id:
                        self.warn_log.write(
                            "QUALIFIER edge at line [{}] associated of corrupted statement edge of id [{}] dropped.\n".format(
                                line_number, self.corrupted_statement_id
                            )
                        )
                        return
        if prop in self.label_set:
            success = self.generate_label_triple(node1, node2)
        elif prop in self.description_set:
            success = self.generate_description_triple(node1, node2)
        elif prop in self.alias_set:
            success = self.generate_alias_triple(node1, node2)
        elif prop == "data_type":
            # special edge of prop declaration
            success = self.generate_prop_declaration_triple(
                node1, node2)
        else:
            if prop in self.prop_types:
                success = self.generate_normal_triple(
                    node1, prop, node2, is_qualifier_edge, e_id)
            else:
                raise KGTKException(
                    "property [{}]'s type is unknown at line [{}].\n".format(
                        prop, line_number)
                )
        if (not success) and self.warning:
            if not is_qualifier_edge: 
                self.warn_log.write(
                        "CORRUPTED_STATEMENT edge at line: [{}] with edge id [{}].\n".format(
                            line_number, e_id))
                self.corrupted_statement_id = e_id
            else:
                self.warn_log.write(
                        "CORRUPTED_QUALIFIER edge at line: [{}] with edge id [{}].\n".format(
                            line_number, e_id))                    
            
        else:
            self.read_num_of_lines += 1
            if not is_qualifier_edge:
                self.to_append_statement_id = e_id
    @staticmethod
    def xsd_number_type(num):
        if isinstance(num, float) and 'e' in str(num).lower():
            return LiteralType.double
        return LiteralType.decimal


class JsonGenerator(Generator):
    def __init__(self,**kwargs):
        super().__init__(**kwargs)
        prop_file = kwargs.pop("prop_file")
        self.output_prefix = kwargs.pop("output_prefix")
        self.file_num = 0
        self.set_properties(prop_file)
        # curret dictionaries
        self.set_json_dict()

    def entry_point(self,line_number, edge):
        edge_list = edge.strip("\n").split("\t")
        l = len(edge_list)
        if line_number == 1:
            # initialize the order_map
            self.initialize_order_map(edge_list)
            return
        node1 = edge_list[self.order_map["node1"]].strip()
        node2 = edge_list[self.order_map["node2"]].strip()
        prop = edge_list[self.order_map["prop"]].strip()
        e_id = edge_list[self.order_map["id"]].strip()

        # add qualifier logic
        if line_number == 2:
            is_qualifier_edge = False
        else:
            if node1 != self.to_append_statement_id and node1 != self.corrupted_statement_id:
                is_qualifier_edge = False
            else:
                is_qualifier_edge = True
                if node1 == self.corrupted_statement_id:
                    if self.warning:
                        self.warn_log.write("QUALIFIER edge at line [{}] associated with corrupted statement edge of id [{}] dropped.\n".format(line_number, self.corrupted_statement_id)
                        )

        # update info_json_dict
        if node1 in self.prop_types:
            success = self.update_misc_json_dict_info(node1, self.prop_types[node1])
        else:
            success = self.update_misc_json_dict_info(node1, None)
        
        assert(success)
        
        if prop in self.prop_types:
            success = self.update_misc_json_dict_info(prop,self.prop_types[prop])
            assert(success)
            if self.prop_types[prop] == "wikibase-item":
                success = self.update_misc_json_dict_info(node2)
                assert(success)
        
        # update label_json_dict
        if prop in self.label_set:
            success = self.update_misc_json_dict_label(node1, prop, node2)
            assert(success)
            return
        
        # update alias and descriptions
        if prop in self.description_set:
            success = self.update_misc_json_dict(node1, prop, node2, line_number,"description")
            assert(success)
            return

        if prop in self.alias_set:
            success = self.update_misc_json_dict(node1, prop, node2, line_number,"alias")
            assert(success)
            return
        
        # normal update for claims & qualifiers
        if is_qualifier_edge:
            success = self.update_misc_json_dict(node1,prop,node2,line_number,"qualifier")
        else:
            success = self.update_misc_json_dict(node1,prop,node2,line_number,"statement")
        
        if (not success) and self.warning:
            if not is_qualifier_edge:
                self.warn_log.write(
                    "CORRUPTED_STATEMENT edge at line: [{}] with edge id [{}].\n".format(
                        line_number, e_id))
                self.corrupted_statement_id = e_id
            else:
                self.warn_log.write(
                        "CORRUPTED_QUALIFIER edge at line: [{}] with edge id [{}].\n".format(
                            line_number, e_id))      
        else:
            # success
            if not is_qualifier_edge:
                self.to_append_statement_id = e_id
                self.to_append_statement = [node1, prop]# path for adding future qualifiers

    def init_entity_in_json(self,node:str):
        self.misc_json_dict[node] = {}
        self.misc_json_dict[node]["labels"] = {}
        self.misc_json_dict[node]["descriptions"] = {}
        self.misc_json_dict[node]["aliases"] = {}
        self.misc_json_dict[node]["claims"] = {}
        self.misc_json_dict[node]["sitelinks"] = {}

    def update_misc_json_dict_label(self,node1:str, prop:str, node2:str):
        if node1 not in self.misc_json_dict:
            self.init_entity_in_json(node1)
        temp_dict = {}
        if node1 not in self.prop_types:
            label_type = "item"
        else:
            label_type = "property"
            label_datatype = self.prop_types[node1]
            temp_dict["datatype"] = label_datatype
        temp_dict["type"] = label_type

        temp_dict["id"] = node1
        temp_dict["labels"] = {}
        if node2 != None:
            text_string, lang = JsonGenerator.process_text_string(node2)
            temp_dict["labels"][lang] = {"language":lang, "value": text_string}
        self.misc_json_dict[node1].update(temp_dict)  

        return True

    def update_misc_json_dict_info(self, node:str,data_type = None):
        if node not in self.misc_json_dict:
            self.init_entity_in_json(node)

        if node.startswith("Q"):
            self.misc_json_dict[node].update(
                {
                "pageid":-1,
                "ns":-1,
                "title":node,
                "lastrevid":"2000-01-01T00:00:00Z", 
                "type":"item",
                "id":node}
                )
        elif node.startswith("P"):
            self.misc_json_dict[node].update(
                {
                "pageid":-1,
                "ns":-1,
                "title":"Property:"+node,
                "lastrevid":"2000-01-01T00:00:00Z",
                "type":"property",
                "datatype":data_type,
                "id":node}
                )
        else:
            raise KGTKException("node {} is neither an entity nor a property.".format(node)) 
        return True
    def update_misc_json_dict(self, node1:str, prop:str, node2:str, line_number:int, field:str):
        if node1 not in self.misc_json_dict:
            self.init_entity_in_json(node1)
        
        if field == "description":
            description_text, lang = JsonGenerator.process_text_string(node2)
            temp_des_dict = {lang:{"languange":lang,"value":description_text}}
            self.misc_json_dict[node1]["descriptions"].update(temp_des_dict)
            return True
        
        if field == "alias":
            alias_text, lang = JsonGenerator.process_text_string(node2)
            temp_alias_dict = {lang, {"languange": lang, "value":alias_text}}
            if lang in self.misc_json_dict[node1]["aliases"]:
                self.misc_json_dict[node1]["aliases"][lang].append(temp_alias_dict)
            else:
                self.misc_json_dict[node1]["aliases"][lang] = [temp_alias_dict]
            return True

        if field == "statement":
            is_qualifier_edge = False
        elif field == "qualifier":
            is_qualifier_edge = True
        
        if prop not in self.prop_types:
            raise KGTKException("property {} at line {} is not defined.".format(prop,line_number))
        
        if prop not in self.misc_json_dict[node1]["claims"]:
                self.misc_json_dict[node1]["claims"][prop] = []
        
        try:
            if self.prop_types[prop] == "wikibase-item":
                object = self.update_misc_json_dict_item(node1, prop, node2, is_qualifier_edge)
            elif self.prop_types[prop] == "time":
                object = self.update_misc_json_dict_time(node1,prop,node2,is_qualifier_edge)
            elif self.prop_types[prop] == "globe-coordinate":
                object = self.update_misc_json_dict_coordinate(node1,prop,node2,is_qualifier_edge)
            elif self.prop_types[prop] == "quantity":
                object = self.update_misc_json_dict_quantity(node1,prop,node2,is_qualifier_edge)
            elif self.prop_types[prop] == "monolingualtext":
                object = self.update_misc_json_dict_monolingualtext(node1,prop,node2,is_qualifier_edge)
            elif self.prop_types[prop] == "string":
                object = self.update_misc_json_dict_string(node1,prop,node2,is_qualifier_edge)
            elif self.prop_types[prop] == "external-id":
                object = self.update_misc_json_dict_external_id(node1,prop,node2,is_qualifier_edge)
            elif self.prop_types[prop] == "url":
                object = self.update_misc_json_dict_url(node1,prop,node2,is_qualifier_edge)
            else:
                raise KGTKException("property tyepe {} of property {} at line {} is not defined.".format(self.prop_types[prop],prop,line_number)) 
            
            if not object:
                if self.warning:
                    self.warn_log.write("edge creation error at line [{}].\n".format(line_number))
                return False

            # process object
            if is_qualifier_edge:
                # update qualifier edge
                if prop in self.misc_json_dict[self.to_append_statement[0]]["claims"][self.to_append_statement[1]][-1]["qualifiers"]:
                    self.misc_json_dict[self.to_append_statement[0]]["claims"][self.to_append_statement[1]][-1]["qualifiers"][prop].append(object)
                else:
                    self.misc_json_dict[self.to_append_statement[0]]["claims"][self.to_append_statement[1]][-1]["qualifiers"][prop] = [object]
                
                # update qualifier order
                if prop not in (self.misc_json_dict[self.to_append_statement[0]]["claims"][self.to_append_statement[1]][-1]["qualifiers-order"]):

                    self.misc_json_dict[self.to_append_statement[0]]["claims"][self.to_append_statement[1]][-1]["qualifiers-order"].append(prop)
            else:
                self.misc_json_dict[node1]["claims"][prop].append(object)
            
            return True

        except:
            raise KGTKException("illegal edge at line {}.".format(line_number))
    
    def update_misc_json_dict_item(self,node1:str,prop:str,node2:str, is_qualifier_edge:bool):
        if not is_qualifier_edge:
            temp_item_dict = {
                    "mainsnak":{
                        "snaktype":"value",
                        "property":prop,
                        "hash":"hashplaceholder",
                        "datavalue":{
                            "value":{
                                "entity-type":"item","numeric-id":0,"id":node2 # place holder for numeric id
                            },
                            "type":"wikibase-entityid"
                        },
                        "datatype":"wikibase-item"
                    },
                    "type":"statement",
                    "id":"id-place-holder",
                    "rank":"normal", #TODO
                    "references":[],
                    "qualifiers":{},
                    "qualifiers-order":[]
                }       
        else:
            temp_item_dict = {
                        "snaktype":"value",
                        "property":prop,
                        "hash":"hashplaceholder",
                        "datavalue":{
                            "value":{
                                "entity-type":"item","numeric-id":0,"id":node2 # place holder for numeric id
                            },
                            "type":"wikibase-entityid"
                        },
                        "datatype":"wikibase-item"
                    }
        return temp_item_dict


    def update_misc_json_dict_time(self,node1:str,prop:str,node2:str,is_qualifier_edge:bool):
        if self.yyyy_pattern.match(node2):
            time_string = node2 + "-01-01"
            precision = 9
        elif self.yyyy_mm_dd_pattern.match(node2):
            time_string = node2
            precision = 11
        try:
            time_string, precision = node2.split("/")
            precision = int(precision)
        except:
            return None
        if not is_qualifier_edge:
            temp_time_dict = {
                "mainsnak":{
                    "snaktype":"value",
                    "property":prop,
                    "hash":"hashplaceholder",
                    "datavalue":{
                        "value":{
                            "time":time_string,
                            "timezone": 0,
                            "before": 0,
                            "after": 0,
                            "precision": precision,
                            "calendarmodel": "http://www.wikidata.org/entity/Q1985727"    
                        },
                        "type":"time"
                    },
                    "datatype":"time"
                },
                "type":"statement",
                "id":"id-place-holder",
                "rank":"normal", #TODO
                "references":[],
                "qualifiers":{},
                "qualifiers-order":[]
                }
        else:
            temp_time_dict = {
                    "snaktype":"value",
                    "property":prop,
                    "hash":"hashplaceholder",
                    "datavalue":{
                        "value":{
                            "time":time_string,
                            "timezone": 0,
                            "before": 0,
                            "after": 0,
                            "precision": precision,
                            "calendarmodel": "http://www.wikidata.org/entity/Q1985727"    
                        },
                        "type":"time"
                    },
                    "datatype":"time"
                }
        return temp_time_dict

    def update_misc_json_dict_coordinate(self,node1:str,prop:str,node2:str,is_qualifier_edge:bool):
        try:
            latitude, longitude = node2[1:].split("/")
            latitude = float(latitude)
            longitude = float(longitude)
        except:
            return None
        if not is_qualifier_edge:
            temp_coordinate_dict = {
                "mainsnak":{
                    "snaktype":"value",
                    "property":prop,
                    "hash":"hashplaceholder",
                    "datavalue":{
                        "value":{
                            "latitude":latitude,
                            "longitude": longitude,
                            "altitude": None,
                            "precision": 0.00027777777777778, # TODO
                            "globe": "http://www.wikidata.org/entity/Q2"    
                        },
                        "type":"globecoordinate"
                    },
                    "datatype":"globecoordinate"
                },
                "type":"statement",
                "id":"id-place-holder",
                "rank":"normal", #TODO
                "references":[],
                "qualifiers":{},
                "qualifiers-order":[]
                }
        else:
            temp_coordinate_dict = {
                    "snaktype":"value",
                    "property":prop,
                    "hash":"hashplaceholder",
                    "datavalue":{
                        "value":{
                            "latitude":latitude,
                            "longitude": longitude,
                            "altitude": None,
                            "precision": 0.00027777777777778, # TODO
                            "globe": "http://www.wikidata.org/entity/Q2"    
                        },
                        "type":"globecoordinate"
                    },
                    "datatype":"globecoordinate"
            }
        return temp_coordinate_dict

    def update_misc_json_dict_quantity(self,node1:str,prop:str,node2:str,is_qualifier_edge:bool):
        try:
            res = self.quantity_pattern.match(node2).groups()
            amount, lower_bound, upper_bound, unit = res
            amount = JsonGenerator.clean_number_string(amount)
            lower_bound = JsonGenerator.clean_number_string(lower_bound)
            upper_bound = JsonGenerator.clean_number_string(upper_bound)
            unit = "http://www.wikidata.org/entity/" + unit if unit != None else None
        except:
            return None
        if not is_qualifier_edge:
            temp_quantity_dict = {
                "mainsnak":{
                    "snaktype":"value",
                    "property":prop,
                    "hash":"hashplaceholder",
                    "datavalue":{
                        "value":{
                            "amount":amount,
                            "unit": unit,  
                            "lowerBound":lower_bound,
                            "UpperBound":upper_bound 
                        },
                        "type":"quantity"
                    },
                    "datatype":"quantity"
                },
                "type":"statement",
                "id":"id-place-holder",
                "rank":"normal", #TODO
                "references":[],
                "qualifiers":{},
                "qualifiers-order":[]
                }
        else:
            temp_quantity_dict = {
                    "snaktype":"value",
                    "property":prop,
                    "hash":"hashplaceholder",
                    "datavalue":{
                        "value":{
                            "amount":amount,
                            "unit": unit,  
                            "lowerBound":lower_bound,
                            "UpperBound":upper_bound 
                        },
                        "type":"quantity"
                    },
                    "datatype":"quantity"
                }
        return temp_quantity_dict
  
    def update_misc_json_dict_monolingualtext(self,node1:str,prop:str,node2:str,is_qualifier_edge:bool):
        text_string, lang = JsonGenerator.process_text_string(node2)
        if not is_qualifier_edge:
            temp_mono_dict ={
                    "mainsnak":{
                        "snaktype":"value",
                        "property":prop,
                        "hash":"hashplaceholder",
                        "datavalue":{
                            "value":{
                                "text":text_string,
                                "language":lang
                            },
                            "type":"monolingualtext"
                        },
                        "datatype":"monolingualtext"
                    },
                    "type":"statement",
                    "id":"id-place-holder",
                    "rank":"normal", #TODO
                    "references":[],
                    "qualifiers":{},
                    "qualifiers-order":[]
                    }
        else:
            temp_mono_dict = {
                        "snaktype":"value",
                        "property":prop,
                        "hash":"hashplaceholder",
                        "datavalue":{
                            "value":{
                                "text":text_string,
                                "language":lang
                            },
                            "type":"monolingualtext"
                        },
                        "datatype":"monolingualtext"
                    }
        return temp_mono_dict
 
    def update_misc_json_dict_string(self,node1:str,prop:str,node2:str,is_qualifier_edge:bool):
        string, lang = JsonGenerator.process_text_string(node2)
        if not is_qualifier_edge:
            temp_string_dict = {
                "mainsnak": {
                "snaktype": "value",
                "property": prop,
                "hash": "hashplaceholder",
                "datavalue": { "value": string, "type": "string" },
                "datatype": "string"
                },
                "type": "statement",
                "id": "id-place-holder",
                "rank": "normal",
                "references":[],
                "qualifiers":{},
                "qualifiers-order":[]
                }
        else:
            temp_string_dict = {
                "snaktype": "value",
                "property": prop,
                "hash": "hashplaceholder",
                "datavalue": { "value": string, "type": "string" },
                "datatype": "string"
                }
        return temp_string_dict

    def update_misc_json_dict_external_id(self,node1:str, prop:str ,node2:str, is_qualifier_edge: bool):

        if not is_qualifier_edge: 
            temp_e_id_dict = {"mainsnak": {
                "snaktype": "value",
                "property": prop,
                "hash": "hashplaceholder",
                "datavalue": { "value": node2, "type": "string" },
                "datatype": "external-id"
            },
            "type": "statement",
            "id": "id-place-holder",
            "rank": "normal",            
            "references":[],
            "qualifiers":{},
            "qualifiers-order":[]
            }
        else:
            temp_e_id_dict = {
                "snaktype": "value",
                "property": prop,
                "hash": "hashplaceholder",
                "datavalue": { "value": node2, "type": "string" },
                "datatype": "external-id"
            }
        return temp_e_id_dict

    def update_misc_json_dict_url(self,node1:str ,prop:str ,node2: str, is_qualifier_edge: bool):
        if not is_qualifier_edge:
            temp_url_dict ={
            "mainsnak": {
                "snaktype": "value",
                "property": prop,
                "hash": "hashplaceholder",
                "datavalue": {
                "value": node2,
                "type": "string"
                },
                "datatype": "url"
            },
            "type": "statement",
            "id": "id-place-holder",
            "rank": "normal",            
            "references":[],
            "qualifiers":{},
            "qualifiers-order":[]
            }
        else:
            temp_url_dict ={
                "snaktype": "value",
                "property": prop,
                "hash": "hashplaceholder",
                "datavalue": {
                "value": node2,
                "type": "string"
                },
                "datatype": "url"
            }
        return temp_url_dict

    def set_properties(self, prop_file:str):
        datatype_mapping = {
            "item": "wikibase-item",
            "time": "time",
            "globe-coordinate": "globe-coordinate",
            "quantity": "quantity",
            "monolingualtext": "monolingualtext",
            "string": "string",
            "external-identifier": "external-id",
            "url": "url"
        }
        with open(prop_file, "r") as fp:
            props = fp.readlines()
        self.prop_types = {}
        for line in props[1:]:
            node1, _, node2 = line.split("\t")
            try:
                self.prop_types[node1] = datatype_mapping[node2.strip()]
            except:
                raise KGTKException(
                    "DataType {} of node {} is not supported.\n".format(
                        node2, node1
                    )
                )
    def set_json_dict(self):
        self.label_json_dict = {}
        self.misc_json_dict = {}
        self.info_json_dict = {}
    def serialize(self):
        '''
        serialize the dictionaries. 
        '''
        # data are aggregated into one file
        JsonGenerator.merge_dict(self.label_json_dict, self.misc_json_dict)
        JsonGenerator.merge_dict(self.info_json_dict, self.misc_json_dict)
        # update dict and files
        with open(self.output_prefix + ".json","w") as fp:
            json.dump(self.misc_json_dict,fp)
        self.set_json_dict()
    @staticmethod
    def merge_dict(source:dict, target: dict):
        for key, value in source.items():
            if isinstance(value, dict):
                # get node or create one
                node = target.setdefault(key, {})
                JsonGenerator.merge_dict(value, node)
            else:
                target[key] = value
        return target


"""Copy records from the first KGTK file to the output file,
compacting repeated items into | lists.

TODO: Need KgtkWriterOptions
"""

from argparse import Namespace, SUPPRESS
from pathlib import Path
import sys
import typing

from kgtk.cli_argparse import KGTKArgumentParser
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.reshape.kgtkidbuilder import KgtkIdBuilder, KgtkIdBuilderOptions
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

def parser():
    return {
        'help': 'Copy a KGTK file, adding ID values.',
        'description': 'Copy a KGTK file, adding ID values.' +
        '\n\nThe --overwrite-id option can be used to replace existing ID values in the ID column. ' +
        'It does not update instances of the same ID in other columns, such as node1, elsewhere in the file.' +
        '\n\nSeveral ID styles are supported. ' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert compact --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_argument(      "input_kgtk_file", nargs="?", type=Path, default="-",
                              help="The KGTK file to filter. May be omitted or '-' for stdin (default=%(default)s).")

    parser.add_argument("-o", "--output-file", dest="output_kgtk_file", help="The KGTK file to write (default=%(default)s).", type=Path, default="-")

    KgtkIdBuilderOptions.add_arguments(parser, expert=True) # Show all the options.
    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_kgtk_file: typing.Optional[Path],
        output_kgtk_file: typing.Optional[Path],

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from kgtk.exceptions import KGTKException

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    idbuilder_options: KgtkIdBuilderOptions = KgtkIdBuilderOptions.from_dict(kwargs)
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("input: %s" % (str(input_kgtk_file) if input_kgtk_file is not None else "-"), file=error_file)
        print("--output-file=%s" % (str(output_kgtk_file) if output_kgtk_file is not None else "-"), file=error_file)
        idbuilder_options.show(out=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:

        # First create the KgtkReader.  It provides parameters used by the ID
        # column builder. Next, create the ID column builder, which provides a
        # possibly revised list of column names for the KgtkWriter.  Create
        # the KgtkWriter.  Last, process the data stream.

        # Open the input file.
        kr: KgtkReader =  KgtkReader.open(input_kgtk_file,
                                          error_file=error_file,
                                          options=reader_options,
                                          value_options=value_options,
                                          verbose=verbose,
                                          very_verbose=very_verbose,
        )
        
        # Create the ID builder.
        idb: KgtkIdBuilder = KgtkIdBuilder.new(kr, idbuilder_options)

        # Open the output file.
        ew: KgtkWriter = KgtkWriter.open(idb.column_names,
                                         output_kgtk_file,
                                         mode=KgtkWriter.Mode[kr.mode.name],
                                         require_all_columns=True,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=False,
                                         gzip_in_parallel=False,
                                         verbose=verbose,
                                         very_verbose=very_verbose)

        # Process the input file, building IDs.
        idb.process(ew)
        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




import os
import sys
import io
import sh # type: ignore

from kgtk.exceptions import KGTKException
import kgtk.cli.zconcat as zcat


def parser():
    return {
        'help': 'Sort file based on one or more columns'
    }

def add_arguments(parser):
    parser.add_argument('-o', '--out', default=None, action='store', dest='output',
                        help='output file to write to, otherwise output goes to stdout')
    parser.add_argument('-c', '--column', '--columns', default='1', action='store', dest='columns',
                        help="comma-separated list of column names or numbers (1-based) to sort on, defaults to 1")
    parser.add_argument('-r', '--reverse', action='store_true', dest='reverse',
                        help="generate output in reverse sort order")
    parser.add_argument('--tsv', action='store_true', dest='tsv',
                        help="assume tab-separated input (default)")
    parser.add_argument('--csv', action='store_true', dest='csv',
                        help="assume comma-separated input (for non-KGTK files)")
    parser.add_argument('--space', action='store_true', dest='space',
                        help="space-optimized configuration for sorting large files")
    parser.add_argument('--speed', action='store_true', dest='speed',
                        help="speed-optimized configuration for sorting large files")
    parser.add_argument('-X', '--extra', default='', action='store', dest='extra',
                        help="extra options to supply to the sort program")
    parser.add_argument('-dt', '--datatype', default='tsv', action='store', dest='_dt',
                        help="Deprecated: datatype of the input file, e.g., tsv (default) or csv.")
    parser.add_argument('input', metavar='INPUT', nargs='?', action='store',
                        help="input file to sort, if empty or `-' process stdin")


# for now: these might require some more refinement:
space_config = '--compress-program=gzip'
speed_config = '--parallel=4'

column_spec_split_regex = '[, |\t]+'
reserved_name_columns = {
    'id': 1, 'ID': 1, 'node1': 1, 'from': 1, 'subject': 1,
    'label': 2, 'predicate': 2, 'property': 2, 'relation': 2, 'relationship': 2,
    'node2': 3, 'to': 3, 'object': 3,
}

def build_sort_key_spec(header, columns, colsep='\t'):
    """Given a KGTK file `header' line and a user-provided `columns' spec, generate a sequence of
    Unix sort key definitions representative of those columns.  For example, columns=subject,object
    will translate into '-k 1,1 -k 3,3'.  Columns can be specified by the names used in the file
    header line, as 1-based positions, or through the pre-defined positions of reserved names such
    as `subject', etc.  Columns found in the header will override any predefined positions.
    """
    import re
    columns = [c.strip() for c in re.split(column_spec_split_regex, columns.strip())]
    header = [c.strip() for c in header.split(colsep)]
    keys = []
    for col in columns:
        if col == '':
            continue
        index = None
        if col in header:
            index = header.index(col) + 1
        else:
            try:
                index = int(col)
            except:
                pass
        if index is None:
            index = reserved_name_columns.get(col)
        if index is None:
            raise KGTKException('Unknown column: ' + col)
        keys.append('-k %d,%d' % (index, index))
    # special whitespace at the end is used by `wait_for_key_spec' below:
    return ' '.join(keys) + ' \t'
    
def build_command(input=None, output=None, columns='1', colsep='\t', options='', _piped=False, _out_mode='wb'):
    """Build a sort sh command pipe for a single `input'.
    If `_piped' is True, configure the last command to ignore `output' and write to a pipe
    in which case this can be used to feed into the input of another command.
    `_out_mode' controls whether an `output' file will be truncated or appened to.
    """
    input = input or '-'
    output = (not _piped and (output or sys.stdout.buffer)) or None
    outfile = None
    if isinstance(output, str):
        outfile = open(output, _out_mode)
        output = outfile
    in_bufsize, out_bufsize = zcat.get_buf_sizes(output=None, _tty_out=False, _piped=True)

    # we unconditionally run zconcat, even if we have an explict input file, since we need to
    # handle optional compression on stdin, and we also need to split off and parse the header:
    sort_pipe = zcat.build_command_1(input, _piped=True)

    # extracting the column header: this is tricky for the case where input comes from stdin, since we
    # have to (optionally) uncompress, then capture the first line, then forward the rest of the input
    # to sort, but its arguments depend on what we find in the header line, so this requires some
    # inter-process communication between different stages in the sorting pipe:
    buffer = io.StringIO()
    catcmd = sh.cat.bake(_piped=True)
    if catcmd.__name__ == sort_pipe[-1].__name__:
        # last command of the pipe is a plain cat, use it instead:
        catcmd = sort_pipe[-1]
    else:
        sort_pipe.append(catcmd)
    # we use local environment vars to communicate the split-off header and sort keys:
    sort_env = os.environ.copy()
    sort_env['KGTK_HEADER'] = zcat.make_temp_file('kgtk-header.')
    sort_env['KGTK_SORT_KEY_SPEC'] = zcat.make_temp_file('kgtk-sort-key-spec.')

    # define these in here, so we can pass in some process-local variables via closures:
    def record_key_spec(chunk):
        # starting with sh 1.13 it looks like we can get either strings or bytes here;
        # if we get bytes we convert to an identical string using `latin1' encoding:
        if isinstance(chunk, bytes):
            chunk = chunk.decode('latin1')
        buffer.write(chunk)
        header = buffer.getvalue()
        eol = header.find('\n')
        if eol >= 0:
            with open(sort_env['KGTK_HEADER'], 'w') as out:
                out.write(header[0:eol+1])
            # reencode from latin1 to utf8 for header processing:
            header = header[0:eol].encode('latin1').decode(zcat.kgtk_encoding)
            with open(sort_env['KGTK_SORT_KEY_SPEC'], 'w') as out:
                out.write(build_sort_key_spec(header, columns, colsep))
            # this signals to ignore the callback once we are done collecting the header:
            return True

    # this waits as a precondition to sort so the header and key files will be available when it starts:
    def wait_for_key_spec():
        import time
        for i in range(100): # try for at most 5 secs:
            with open(sort_env['KGTK_SORT_KEY_SPEC'], 'r') as inp:
                if inp.read().endswith('\t'):
                    break
                time.sleep(0.05)
        else:
            raise KGTKException('INTERNAL ERROR: failed to communicate sort key')

    # baking in additional args creates a new command, so we have to update the pipe;
    # for some reason output callbacks don't work with _piped, so we use _bg instead:
    catcmd = catcmd.bake(_bg=True, _piped=False, _out=record_key_spec, _tee='out', _in_bufsize=in_bufsize, _out_bufsize=out_bufsize)
    sort_pipe[-1] = catcmd

    # rest of the pipeline: first skip header line, then sort, then add header back in:
    sort_pipe.append(
        sh.tail.bake('-n', '+2', _piped=True, _in_bufsize=in_bufsize, _out_bufsize=out_bufsize)
    )
    sort_pipe.append(
        sh.sh.bake('-c', """exec sort -t "%s" %s `cat $KGTK_SORT_KEY_SPEC`""" % (colsep, options),
                   _env=sort_env, _piped=True, _preexec_fn=wait_for_key_spec, _in_bufsize=in_bufsize, _out_bufsize=out_bufsize)
    )
    in_bufsize, out_bufsize = zcat.get_buf_sizes(output=output, _piped=_piped)
    cleanup = lambda cmd, status, exit_code: [sh.rm('-f', sort_env['KGTK_HEADER']), sh.rm('-f', sort_env['KGTK_SORT_KEY_SPEC'])]
    sort_pipe.append(
        sh.sh.bake('-c', """exec cat $KGTK_HEADER -""",
                   _env=sort_env, _out=output, _done=cleanup, _piped=_piped, _in_bufsize=in_bufsize, _out_bufsize=out_bufsize)
    )
    return sort_pipe


def run(input=None, output=None, columns='1', reverse=False, space=False, speed=False, extra='', tsv=False, csv=False, _dt=None):
    """Run sort according to the provided command-line arguments.
    """
    try:
        colsep = '\t'
        if not tsv and (csv or _dt == 'csv'):
            colsep = ','

        options = extra
        if reverse:
            options += ' -r'
        if space:
            options += ' ' + space_config
        elif speed:
            options += ' ' + speed_config
            
        pipe = build_command(input=input, output=output, columns=columns, colsep=colsep, options=options)
        return zcat.run_sh_commands(pipe).exit_code
    except sh.SignalException_SIGPIPE:
        # hack to work around Python3 issue when stdout is gone when we try to report an exception;
        # without this we get an ugly 'Exception ignored...' msg when we quit with head or a pager:
        sys.stdout = os.fdopen(1)
    except Exception as e:
        #import traceback
        #traceback.print_tb(sys.exc_info()[2], 10)
        raise KGTKException('INTERNAL ERROR: ' + type(e).__module__ + '.' + str(e) + '\n')

"""
# old mlr-based version:
def run(datatype, column, input): 
    sh.mlr('--%s' % datatype, 'sort', '-f', column, input, _out=sys.stdout, _err=sys.stderr)
"""

"""
# Examples:

> cat /tmp/nodes-v2-slice-shuf.tsv.gz | kgtk sort -c 'label, id' | head
id	label	type	descriptions	aliases	document_id
Q28415		item	'railway station'@en		wikidata-20200203
Q45582		item	'Polish literary award'@en		wikidata-20200203
Q45877		item	'television series'@en		wikidata-20200203
Q45886		item			wikidata-20200203
Q46028		item			wikidata-20200203
Q46103		item	'Wikimedia list article'@en		wikidata-20200203
Q46106		item	'Wikimedia list article'@en		wikidata-20200203
Q46430		item			wikidata-20200203
Q46473		item			wikidata-20200203


# uncompress & sort 9GB file from stdin in about 4 min on laptop (the -X option is just for illustration):

> date; cat nodes-v2.csv.gz | kgtk sort -c 'id, label' --speed -X ' -T /data/tmp' -o /tmp/nodes-v2-sorted.tsv; date
Wed 08 Apr 2020 07:25:46 PM PDT
Wed 08 Apr 2020 07:29:37 PM PDT

> ls -l /tmp/nodes-v2-sorted.tsv
-rw-r--r-- 1 hans isdstaff 8918052340 Apr  8 19:29 /tmp/nodes-v2-sorted.tsv

> head -3 /tmp/nodes-v2-sorted.tsv
id	label	type	descriptions	aliases	document_id
P10	'video'@en	property	'relevant video. For images, use the property P18. For film trailers, qualify with 'object has role' (P3831)='trailer' (Q622550)'@en	'media'@en|'animation'@en|'gif'@en|'trailer (Commons)'@en	wikidata-20200203
P1000	'record held'@en	property	'notable record achieved by a person or entity, include qualifiers for dates held'@en		wikidata-20200203
"""



"""
Import an wikidata file into KGTK file
"""

def parser():
    return {
        'help': 'Import an wikidata file into KGTK file'
    }


def add_arguments(parser):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    parser.add_argument("-i",'--inp', action="store", type=str, dest="inp_path",help='input path file')
    parser.add_argument(
        '--procs',
        action="store",
        type=int,
        dest="procs",
        default=2,
        help='number of processes to run in parallel, default 2')
    parser.add_argument(
        "--node",
        action="store",
        type=str,
        dest="node_file",
        default=None,
        help='path to output node file')
    parser.add_argument(
        "--edge",
        action="store",
        type=str,
        dest="edge_file",
        default=None,
        help='path to output edge file')
    parser.add_argument(
        "--qual",
        action="store",
        type=str,
        dest="qual_file",
        default=None,
        help='path to output qualifier file')
    parser.add_argument(
        "--limit",
        action="store",
        type=int,
        dest="limit",
        default=None,
        help='number of lines of input file to run on, default runs on all')
    parser.add_argument(
        "--lang",
        action="store",
        type=str,
        dest="lang",
        default="en",
        help='languages to extract, comma separated, default en')
    parser.add_argument(
        "--source",
        action="store",
        type=str,
        dest="source",
        default="wikidata",
        help='wikidata version number, default: wikidata')
    parser.add_argument(
        "--deprecated",
        action="store_true",
        dest="deprecated",
        help='option to include deprecated statements, not included by default')
    
    
def run(inp_path,procs,node_file,edge_file,qual_file,limit,lang,source,deprecated):
    # import modules locally
    import bz2
    import simplejson as json
    import csv
    import pyrallel
    import time
    import sh
    from kgtk.cli_argparse import KGTKArgumentParser
    from kgtk.exceptions import KGTKException

    

    class MyMapper(pyrallel.Mapper):

        def enter(self):
            WD_META_ITEMS = [
                "Q163875",
                "Q191780",
                "Q224414",
                "Q4167836",
                "Q4167410",
                "Q4663903",
                "Q11266439",
                "Q13406463",
                "Q15407973",
                "Q18616576",
                "Q19887878",
                "Q22808320",
                "Q23894233",
                "Q33120876",
                "Q42104522",
                "Q47460393",
                "Q64875536",
                "Q66480449",
            ]
            # filter: currently defined as OR: one hit suffices to be removed from
            # further processing
            exclude_list = WD_META_ITEMS

            # punctuation
            exclude_list.extend(["Q1383557", "Q10617810"])

            # letters etc
            exclude_list.extend(["Q188725", "Q19776628", "Q3841820",
                                 "Q17907810", "Q9788", "Q9398093"])

            self.neg_prop_filter = {
                'P31': exclude_list,    # instance of
                'P279': exclude_list    # subclass
            }
            self.parse_labels=True
            self.parse_aliases=True
            self.parse_descr=True
            self.first=True
            self.cnt=0
            self.write_mode='w'

        def process(self,line,node_file,edge_file,qual_file,languages,doc_id):
            write_mode='a'
            if self.first==True:
                write_mode='w'
                self.first=False
            if self.cnt % 500000 == 0 and self.cnt>0:
                print("{} lines processed by processor {}".format(self.cnt,self._idx))
            self.cnt+=1
            nrows=[]
            erows=[]
            qrows=[]
            clean_line = line.strip()
            if clean_line.endswith(b","):
                clean_line = clean_line[:-1]
            if len(clean_line) > 1:
                obj = json.loads(clean_line)
                entry_type = obj["type"]
                if entry_type == "item" or entry_type == "property":
                    keep = True
                if node_file and keep:
                    row = []
                    qnode = obj["id"]
                    row.append(qnode)

                    if self.parse_labels:
                        labels = obj["labels"]
                        label_list=[]
                        if labels:
                            for lang in languages:
                                lang_label = labels.get(lang, None)
                                if lang_label:
                                    lang_label['value']=lang_label['value'].replace('|','\\|')
                                    label_list.append(
                                        '\'' + lang_label['value'].replace("'","\\'") + '\'' + "@" + lang)
                        if len(label_list)>0:
                            row.append("|".join(label_list))
                        else:
                            row.append("")
                    row.append(entry_type)

                    if self.parse_descr:
                        descriptions = obj["descriptions"]
                        descr_list=[]
                        if descriptions:
                            for lang in languages:
                                lang_descr = descriptions.get(lang, None)
                                if lang_descr:
                                    lang_descr['value']=lang_descr['value'].replace('|','\\|')
                                    descr_list.append(
                                        '\'' + lang_descr['value'].replace("'","\\'") + '\'' + "@" + lang)
                        if len(descr_list)>0:
                            row.append("|".join(descr_list))
                        else:
                            row.append("")

                    if self.parse_aliases:
                        aliases = obj["aliases"]
                        alias_list = []
                        if aliases:
                            for lang in languages:
                                lang_aliases = aliases.get(lang, None)
                                if lang_aliases:
                                    for item in lang_aliases:
                                        item['value']=item['value'].replace('|','\\|')
                                        alias_list.append(
                                            '\'' + item['value'].replace("'","\\'") + '\'' + "@" + lang)
                        if len(alias_list)>0:
                            row.append("|".join(alias_list))
                        else:
                            row.append("")

                    #row.append(doc_id)
                    if node_file:
                        nrows.append(row)

                if edge_file or qual_file:
                    claims = obj["claims"]
                    for prop, value_set in self.neg_prop_filter.items():
                        claim_property = claims.get(prop, None)
                        if claim_property:
                            for cp in claim_property:
                                cp_id = (
                                    cp["mainsnak"]
                                    .get("datavalue", {})
                                    .get("value", {})
                                    .get("id")
                                )
                                cp_rank = cp["rank"]
                                if cp_rank != "deprecated" and cp_id in value_set:
                                    keep = False
                    if keep:
                        sitelinks=obj.get('sitelinks',None)
                        qnode = obj["id"]
                        for prop, claim_property in claims.items():
                            seq_no = 1
                            for cp in claim_property:
                                if (deprecated or cp['rank'] != 'deprecated') and cp['mainsnak']['snaktype'] == 'value':
                                    rank=cp['rank']
                                    val = cp['mainsnak']['datavalue'].get(
                                        'value')
                                    typ = cp['mainsnak']['datatype']
                                    sid = qnode + '-' + \
                                        prop + '-' + str(seq_no)                             
                                    value = ''
                                    mag = ''
                                    unit = ''
                                    date=''
                                    item=''
                                    lower = ''
                                    upper = ''
                                    precision = ''
                                    calendar = ''
                                    lat = ''
                                    long = ''
                                    enttype = ''
                                    if typ.startswith('wikibase'):
                                        enttype = val.get('entity-type')
                                        value = val.get('id', '')
                                        item=value
                                    elif typ == 'quantity':
                                        value = val['amount']
                                        mag = val['amount']
                                        if val.get(
                                                'upperBound',
                                                None) or val.get(
                                                'lowerBound',
                                                None):
                                            lower = val.get('lowerBound', '')
                                            upper = val.get('upperBound', '')
                                            value += '[' + lower + \
                                                ',' + upper + ']'
                                        if len(val.get('unit')) > 1:
                                            unit = val.get(
                                                'unit').split('/')[-1]
                                            value += unit
                                    elif typ == 'globe-coordinate':
                                        lat = str(val['latitude'])
                                        long = str(val['longitude'])
                                        precision = val.get('precision', '')
                                        value = '@' + lat + '/' + long
                                    elif typ == 'time':
                                        if val['time'][0]=='-':
                                            pre="^-"
                                        else:
                                            pre="^"
                                        date = pre + val['time'][1:]
                                        precision = str(val['precision'])
                                        calendar = val.get(
                                            'calendarmodel', '').split('/')[-1]
                                        value = pre + \
                                            val['time'][1:] + '/' + str(val['precision'])
                                    elif typ == 'monolingualtext':
                                        value = '\'' + \
                                            val['text'].replace("'","\\'") + '\'' + '@' + val['language']
                                    else:
                                        value = '\"' + val.replace('"','\\"') + '\"'
                                    if edge_file:
                                        erows.append([sid,
                                                     qnode,
                                                     prop,
                                                     value,
                                                     rank,
                                                     mag,
                                                     unit,
                                                     date,
                                                     item,
                                                     lower,
                                                     upper,
                                                     lat,
                                                     long,
                                                     precision,
                                                     calendar,
                                                     enttype])
                                    seq_no += 1
                                    if qual_file:
                                        if cp.get('qualifiers', None):
                                            quals = cp['qualifiers']
                                            for qual_prop, qual_claim_property in quals.items():
                                                qual_seq_no = 1
                                                for qcp in qual_claim_property:
                                                    if qcp['snaktype'] == 'value':
                                                        value = ''
                                                        mag = ''
                                                        unit = ''
                                                        date= ''
                                                        item=''
                                                        lower = ''
                                                        upper = ''
                                                        precision = ''
                                                        calendar = ''
                                                        lat = ''
                                                        long = ''
                                                        enttype = ''
                                                        val = qcp['datavalue'].get(
                                                            'value')
                                                        typ = qcp['datatype']
                                                        tempid = sid + '-' + qual_prop + \
                                                            '-' + str(qual_seq_no)
                                                        qual_seq_no += 1
                                                        if typ.startswith(
                                                                'wikibase'):
                                                            enttype = val.get(
                                                                'entity-type')
                                                            value = val.get(
                                                                'id', '')
                                                            item=value
                                                        elif typ == 'quantity':
                                                            value = val['amount']
                                                            mag = val['amount']
                                                            if val.get(
                                                                    'upperBound',
                                                                    None) or val.get(
                                                                    'lowerBound',
                                                                    None):
                                                                lower = val.get(
                                                                    'lowerBound', '')
                                                                upper = val.get(
                                                                    'upperBound', '')
                                                                value += '[' + lower + \
                                                                    ',' + upper + ']'
                                                            if len(
                                                                    val.get('unit')) > 1:
                                                                unit = val.get(
                                                                    'unit').split('/')[-1]
                                                                value += unit
                                                        elif typ == 'globe-coordinate':
                                                            lat = str(
                                                                val['latitude'])
                                                            long = str(
                                                                val['longitude'])
                                                            precision = val.get(
                                                                'precision', '')
                                                            value = '@' + lat + '/' + long
                                                        elif typ == 'time':
                                                            if val['time'][0]=='-':
                                                                pre="^-"
                                                            else:
                                                                pre="^"
                                                            date = pre + \
                                                                val['time'][1:]
                                                            precision = str(
                                                                val['precision'])
                                                            calendar = val.get(
                                                                'calendarmodel', '').split('/')[-1]
                                                            value = pre + \
                                                                val['time'][1:] + '/' + str(val['precision'])
                                                        elif typ == 'monolingualtext':
                                                            value = '\'' + \
                                                                val['text'].replace("'","\\'") + '\'' + '@' + val['language']
                                                        else:
                                                            value = '\"' + val.replace('"','\\"') + '\"'
                                                        qrows.append(
                                                            [
                                                                tempid,
                                                                sid,
                                                                qual_prop,
                                                                value,
                                                                mag,
                                                                unit,
                                                                date,
                                                                item,
                                                                lower,
                                                                upper,
                                                                lat,
                                                                long,
                                                                precision,
                                                                calendar,
                                                                enttype])
            
                        if sitelinks:
                            wikipedia_seq_no = 1
                            for link in sitelinks:
                                if link.endswith('wiki') and link!='commonswiki':
                                    sid=qnode + '-wikipedia_sitelink-'+str(wikipedia_seq_no)
                                    wikipedia_seq_no+=1
                                    sitetitle='_'.join(sitelinks[link]['title'].split())
                                    sitelang=link.split('wiki')[0].replace('_','-')
                                    sitelink='http://'+sitelang+'.wikipedia.org/wiki/'+sitetitle
                                    if edge_file:
                                        erows.append([sid, qnode, 'wikipedia_sitelink', sitelink,'','','','','','','',
                                                      '','','','',''])
                                    if qual_file:
                                        tempid=sid+'-language-1'
                                        qrows.append([tempid,sid,'language',sitelang,'','','','','','','','','','',''])

            if node_file:
                with open(node_file+'_{}'.format(self._idx), write_mode, newline='') as myfile:
                    for row in nrows:
                        wr = csv.writer(
                            myfile,
                            quoting=csv.QUOTE_NONE,
                            delimiter="\t",
                            escapechar="\n",
                            quotechar='')
                        wr.writerow(row)
            if edge_file:
                with open(edge_file+'_{}'.format(self._idx), write_mode, newline='') as myfile:
                    for row in erows:
                        wr = csv.writer(
                            myfile,
                            quoting=csv.QUOTE_NONE,
                            delimiter="\t",
                            escapechar="\n",
                            quotechar='')
                        wr.writerow(row)
            if qual_file:
                with open(qual_file+'_{}'.format(self._idx), write_mode, newline='') as myfile:
                    for row in qrows:
                        wr = csv.writer(
                            myfile,
                            quoting=csv.QUOTE_NONE,
                            delimiter="\t",
                            escapechar="\n",
                            quotechar='')
                        wr.writerow(row)
    

    
    try:
        start=time.time()
        languages=lang.split(',')
        if node_file:
            header = ['id','label','type','description','alias']
            with open(node_file+'_header', 'w', newline='') as myfile:
                wr = csv.writer(
                    myfile,
                    quoting=csv.QUOTE_NONE,
                    delimiter="\t",
                    escapechar="\n",
                    quotechar='')
                wr.writerow(header)
        header = ['id','node1','label','node2','rank','node2;magnitude','node2;unit','node2;date','node2;item','node2;lower','node2;upper',
              'node2;latitude','node2;longitude','node2;precision','node2;calendar','node2;entity-type']
        if edge_file:
            with open(edge_file+'_header', 'w', newline='') as myfile:
                wr = csv.writer(
                    myfile,
                    quoting=csv.QUOTE_NONE,
                    delimiter="\t",
                    escapechar="\n",
                    quotechar='')
                wr.writerow(header)
        if qual_file:
            header.remove('rank')
            with open(qual_file+'_header', 'w', newline='') as myfile:
                wr = csv.writer(
                    myfile,
                    quoting=csv.QUOTE_NONE,
                    delimiter="\t",
                    escapechar="\n",
                    quotechar='')
                wr.writerow(header)
        pp = pyrallel.ParallelProcessor(procs, MyMapper,enable_process_id=True)
        pp.start()
        with bz2.open(inp_path, mode='rb') as file:
            print('processing wikidata file now...')
            for cnt, line in enumerate(file):
                if limit and cnt >= limit:
                    break
                pp.add_task(line,node_file,edge_file,qual_file,languages,source)
        pp.task_done()
        pp.join()
        if node_file:
            cat_command=[node_file+'_header']
            rm_command=[node_file+'_header']
            for n in range(procs):
                cat_command.append(node_file+'_'+str(n))
                rm_command.append(node_file+'_'+str(n))
            if limit and limit==1:
                sh.cat(node_file+'_header',node_file+'_0',_out=node_file)
                sh.rm(node_file+'_header',node_file+'_0') 
            else:
                sh.cat(*cat_command,_out=node_file)
                sh.rm(*rm_command)
        if edge_file:
            cat_command=[edge_file+'_header']
            rm_command=[edge_file+'_header']
            for n in range(procs):
                cat_command.append(edge_file+'_'+str(n))
                rm_command.append(edge_file+'_'+str(n))
            if limit and limit==1:
                sh.cat(edge_file+'_header',edge_file+'_0',_out=edge_file)
                sh.rm(edge_file+'_header',edge_file+'_0') 
            else:
                sh.cat(*cat_command,_out=edge_file)
                sh.rm(*rm_command)
        if qual_file:
            cat_command=[qual_file+'_header']
            rm_command=[qual_file+'_header']
            for n in range(procs):
                cat_command.append(qual_file+'_'+str(n))
                rm_command.append(qual_file+'_'+str(n))
            if limit and limit==1:
                sh.cat(qual_file+'_header',qual_file+'_0',_out=qual_file)
                sh.rm(qual_file+'_header',qual_file+'_0') 
            else:
                sh.cat(*cat_command,_out=qual_file)
                sh.rm(*rm_command)
        print('import complete')
        end=time.time()
        print('time taken : {}s'.format(end-start))
    except:
        raise KGTKException



"""Count the unique values in a column in an KGTK file.  Generate an output
KGTK node or edge file with the counts.  Empty values are omitted from the
output KGTK file unless a non-empty substitute value is provided.

TODO: Consider other output formats. Perhaps seperate counts for each node1
(node2, etc.) value in the input file?

"""

from argparse import ArgumentParser, Namespace
import attr
from pathlib import Path
import sys
import typing

from kgtk.kgtkformat import KgtkFormat
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

@attr.s(slots=True, frozen=True)
class Unique(KgtkFormat):
    input_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))

    column_name: str = attr.ib(validator=attr.validators.instance_of(str))

    output_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))

    empty_value: str = attr.ib(validator=attr.validators.instance_of(str), default="")

    label_value: str = attr.ib(validator=attr.validators.instance_of(str), default="count")

    # TODO: make this an enum
    output_format: str = attr.ib(validator=attr.validators.instance_of(str), default="edge")
    prefix: str = attr.ib(validator=attr.validators.instance_of(str), default="")

    # TODO: find working validators
    # value_options: typing.Optional[KgtkValueOptions] = attr.ib(attr.validators.optional(attr.validators.instance_of(KgtkValueOptions)), default=None)
    reader_options: typing.Optional[KgtkReaderOptions]= attr.ib(default=None)
    value_options: typing.Optional[KgtkValueOptions] = attr.ib(default=None)

    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    very_verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    def process(self):
        # Open the input file.
        if self.verbose:
            if self.input_file_path is not None:
                print("Opening the input file: %s" % self.input_file_path, file=self.error_file, flush=True)
            else:
                print("Reading the input data from stdin", file=self.error_file, flush=True)

        kr: KgtkReader =  KgtkReader.open(self.input_file_path,
                                          error_file=self.error_file,
                                          options=self.reader_options,
                                          value_options = self.value_options,
                                          verbose=self.verbose,
                                          very_verbose=self.very_verbose,
        )

        if self.column_name not in kr.column_name_map:
            raise ValueError("Column %s is not in the input file" % (self.column_name))
        column_idx: int = kr.column_name_map[self.column_name]

        if self.verbose:
            print("Counting unique values from the %s column in %s" % (self.column_name, self.input_file_path), file=self.error_file, flush=True)
        input_line_count: int = 0

        value_counts: typing.MutableMapping[str, int] = { }
        
        row: typing.list[str]
        for row in kr:
            input_line_count += 1
            value: str = row[column_idx]
            if len(value) == 0:
                value = self.empty_value
            if len(value) > 0:
                value = self.prefix + value
                value_counts[value] = value_counts.get(value, 0) + 1
                
        if self.verbose:
            print("Read %d records, found %d unique non-empty values, %d empty values." % (input_line_count,
                                                                                           len(value_counts),
                                                                                           input_line_count - len(value_counts)),
                  file=self.error_file, flush=True)

        # No node mode we can't open the output file until we are done reading
        # the input file, because we need the list of uniqueue values to
        # build the column list.
        output_columns: typing.List[str]
        if self.output_format == "edge":
            output_columns = ["node1", "label", "node2"]
        elif self.output_format == "node":
            output_columns = [ "id" ]
            for value in sorted(value_counts.keys()):
                # TODO: provide a way to override this check.
                if value in KgtkFormat.NODE1_COLUMN_NAMES:
                    raise ValueError("Cannot write a KGTK node file with a column named '%s'." % value)
                output_columns.append(value)
        else:
            raise ValueError("Unknown output format %s" % str(self.output_format))
        
        if self.verbose:
            print("Opening the output file: %s" % self.output_file_path, file=self.error_file, flush=True)

        ew: KgtkWriter = KgtkWriter.open(output_columns,
                                         self.output_file_path,
                                         require_all_columns=False,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=True,
                                         gzip_in_parallel=False,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose)        

        if self.output_format == "edge":
            for value in sorted(value_counts.keys()):
                ew.write([value, self.label_value, str(value_counts[value])])
        elif self.output_format == "node":
            row = [ self.column_name ]
            for value in sorted(value_counts.keys()):
                row.append(str(value_counts[value]))
            ew.write(row)
        else:
            raise ValueError("Unknown output format %s" % str(self.output_format))

        ew.close()
       
def main():
    """
    Test the KGTK unique processor.
    """
    parser: ArgumentParser = ArgumentParser()

    parser.add_argument(dest="input_file_path", help="The KGTK file with the input data", type=Path, nargs="?")

    parser.add_argument(      "--column", dest="column_name", help="The column to count unique values (required).", required=True)

    parser.add_argument(      "--empty", dest="empty_value", help="A value to substitute for empty values (default=%(default)s).", default="")

    parser.add_argument("-o", "--output-file", dest="output_file_path", help="The KGTK file to write (default=%(default)s).", type=Path, default="-")

    parser.add_argument(      "--label", dest="label_value", help="The output file label column value (default=%(default)s).", default="count")

    # TODO: use an enum
    parser.add_argument(      "--format", dest="output_format", help="The output file format and mode (default=%(default)s).",
                              default="edge", choices=["edge", "node"])

    parser.add_argument(      "--prefix", dest="prefix", help="The value prefix (default=%(default)s).", default="")

    KgtkReader.add_debug_arguments(parser)
    KgtkReaderOptions.add_arguments(parser, mode_options=True)
    KgtkValueOptions.add_arguments(parser)

    args: Namespace = parser.parse_args()

    error_file: typing.TextIO = sys.stdout if args.errors_to_stdout else sys.stderr

    # Build the option structures.                                                                                                                          
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_args(args)
    value_options: KgtkValueOptions = KgtkValueOptions.from_args(args)

   # Show the final option structures for debugging and documentation.                                                                                             
    if args.show_options:
        print("input: %s" % (str(args.input_file_path) if args.input_file_path is not None else "-"), file=error_file)
        print("--column=%s" % args.column_name, file=error_file)
        print("--empty=%s" % args.empty_value, file=error_file)
        print("--output-file=%s" % str(args.output_file_path), file=error_file)
        print("--label=%s" % args.label_value, file=error_file)
        print("--format=%s" % args.output_format, file=error_file)
        print("--prefix=%s" % args.prefix, file=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)

    uniq: Unique = Unique(
        input_file_path=args.input_file_path,
        column_name=args.column_name,
        output_file_path=args.output_file_path,
        empty_value=args.empty_value,
        label_value=args.label_value,
        output_format=args.output_format,
        prefix=args.prefix,
        reader_options=reader_options,
        value_options=value_options,
        error_file=error_file,
        verbose=args.verbose,
        very_verbose=args.very_verbose)

    uniq.process()

if __name__ == "__main__":
    main()



"""
Convert a KGTK file to a GitHub markdown table.

TODO: Need KgtkWriterOptions
"""

from argparse import Namespace, SUPPRESS
from pathlib import Path
import sys
import typing

from kgtk.cli_argparse import KGTKArgumentParser
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.join.kgtkcat import KgtkCat
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

def parser():
    return {
        'help': 'Convert a KGTK file to a GitHub Markdown Table.',
        'description': 'Convert a KGTK input file to a GitHub markdown table on output. ' +
        '\n\nUse this command to filter the output of any KGTK command: ' +
        '\n\nkgtk md ' +
        '\n\nUse it to convert a KGTK file to a GitHub Markdown table in a file: ' +
        '\n\nkgtk md file.tsv ' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert cat --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_argument(      "input_file_path", help="The KGTK file to convert to a GitHub markdown table.", type=Path, nargs='?', default=Path("-"))

    parser.add_argument("-o", "--output-file", dest="output_file_path", help="The KGTK file to write (default=%(default)s).", type=Path, default="-")

    parser.add_argument(      "--output-format", dest="output_format", help=h("The file format (default=%(default)s)"), type=str, default="md")

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file_path: Path,
        output_file_path: Path,
        output_format: str,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from kgtk.exceptions import KGTKException


    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # TODO: check that at most one input file is stdin?

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("input: %s" % str(input_file_path), file=error_file, flush=True)
        print("--output-file=%s" % str(output_file_path), file=error_file, flush=True)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:
        kc: KgtkCat = KgtkCat(input_file_paths=[input_file_path],
                              output_path=output_file_path,
                              output_format=output_format,
                              reader_options=reader_options,
                              value_options=value_options,
                              error_file=error_file,
                              verbose=verbose,
                              very_verbose=very_verbose
        )
        
        kc.process()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




"""
Constants and helpers for the KGTK file format.

"""

from enum import Enum, unique
import sys
import typing

class KgtkFormat:
    COLUMN_SEPARATOR: str = "\t"
    COMMENT_INDICATOR: str = "#"
    LIST_SEPARATOR: str = "|"

    # These are the required columns in an edge file:
    NODE1_COLUMN_NAMES: typing.List[str] = ["node1", "from", "subject"]
    NODE2_COLUMN_NAMES: typing.List[str] = ["node2", "to", "object"]
    LABEL_COLUMN_NAMES: typing.List[str] = ["label", "predicate", "relation", "relationship"]

    # There is only one required column in a node file:
    ID_COLUMN_NAMES: typing.List[str] = ["id", "ID"]

    # These are the canonical names:
    NODE1: str = "node1"
    LABEL: str = "label"
    NODE2: str = "node2"
    ID: str = "id"

    KGTK_NAMESPACE: str = "kgtk:"

    @unique
    class DataType(Enum):
        EMPTY = 0
        LIST = 1
        NUMBER = 2
        QUANTITY = 3
        STRING = 4
        LANGUAGE_QUALIFIED_STRING = 5
        LOCATION_COORDINATES = 6
        DATE_AND_TIMES = 7
        EXTENSION = 8
        BOOLEAN = 9
        SYMBOL = 10

        def lower(self)->str:
            return self.name.lower()

        @classmethod
        def choices(cls)->typing.List[str]:
            results: typing.List[str] = [ ]
            name: str
            for name in cls.__members__.keys():
                results.append(name.lower())
            return results

    TRUE_SYMBOL: str = "True"
    FALSE_SYMBOL: str = "False"



"""
Generate mediawiki API json files from kgtk file

This command line tool will create three json files. Each will mimic the return of the following media wiki API for each entity existed in the kgtk file. 

"""

def str2bool(v):
    import argparse
    if isinstance(v, bool):
       return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

def parser():
    """
    Initialize sub-parser.
    Parameters: https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser
    """
    return {
        "help": "Generates mediawiki json responses from kgtk file",
        "description": "Generating json files that mimic mediawiki *wbgetentities* api call response. This tool assumes statements and qualifiers related to one entity will be bundled close as the `generate_wikidata_triples` function assumes. If this requirement is not met, please set `n` to a number LARGER than the total number of entities in the kgtk file",
    }

def add_arguments(parser):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
        prop_file: str
    """
    parser.add_argument(
        "-lp",
        "--label-property",
        action="store",
        type=str,
        default="label",
        required=False,
        help="property identifiers which will create labels, separated by comma','.",
        dest="labels",
    )
    parser.add_argument(
        "-ap",
        "--alias-property",
        action="store",
        type=str,
        required = False,
        default="alias",
        help="alias identifiers which will create labels, separated by comma','.",
        dest="aliases",
    )
    parser.add_argument(
        "-dp",
        "--description-property",
        action="store",
        type=str,
        required = False,
        default="description",
        help="description identifiers which will create labels, separated by comma','.",
        dest="descriptions",
    )
    parser.add_argument(
        "-pf",
        "--property-file",
        action="store",
        type=str,
        required = True,
        help="path to the file which contains the property datatype mapping in kgtk format.",
        dest="prop_file",
    )
    parser.add_argument(
        "-gz",
        "--use-gz",
        action="store",
        type=str2bool,
        required = False,
        default="no",
        help="if set to yes, read from compressed gz file",
        dest="use_gz",
    )
    parser.add_argument(
        "-pr",
        "--output-file-prefix",
        action="store",
        type=str,
        default = "kgtk",
        required = False,
        help="set the prefix of the output files. Default to `kgtk`",
        dest="output_prefix",
    )
    parser.add_argument(
        "-n",
        "--output-n-lines",
        action="store",
        type=int,
        required = False,
        default=1000,
        help="output json file when the corresponding dictionary size reaches n. Default to 1000",
        dest="n",
    )
    parser.add_argument(
        "-log",
        "--log-path",
        action="store",
        type=str,
        required = False,
        default="warning.log",
        help="set the path of the log file",
        dest="log_path",
    )
    parser.add_argument(
        "-w",
        "--warning",
        action="store",
        type=str2bool,
        required = False,
        default="no",
        help="if set to yes, warn various kinds of exceptions and mistakes and log them to a log file with line number in input file, rather than stopping. logging",
        dest="warning",
    )


def run(
    labels: str,
    aliases: str,
    descriptions: str,
    prop_file: str,
    use_gz: bool,
    output_prefix: str,
    n: int,
    log_path: str,
    warning: bool
):
    # import modules locally
    from kgtk.generator import JsonGenerator
    import sys
    import gzip
    
    generator = JsonGenerator(
        label_set=labels,
        alias_set=aliases,
        description_set=descriptions,
        prop_file=prop_file,
        output_prefix = output_prefix,
        n = n,
        log_path = log_path,
        warning = warning
    )
    # process stdin
    if use_gz:
        fp = gzip.open(sys.stdin.buffer, 'rt')
    else:
        fp = sys.stdin
        # not line by line
    for line_num, edge in enumerate(fp):
        if edge.startswith("#") or len(edge.strip("\n")) == 0:
            continue
        else:
            generator.entry_point(line_num+1,edge)
    generator.finalize()



"""Add label columns for values in the node1, label, and node2 fields.

The input rows are saved in memory, as well as the value-to-label mapping.
This will impose a limit on the size of the input files that can be processed.

TODO: Optionally save the input rows in an external disk file?

TODO: Optionally reread the input stream insted of saving the input rows?
      With special provision for copying standard input to an external disk file.

TODO: Need KgtkWriterOptions

TODO: Provide seperate reader options for the label file.

"""

from argparse import ArgumentParser, Namespace
import attr
from pathlib import Path
import sys
import typing

from kgtk.kgtkformat import KgtkFormat
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalue import KgtkValue
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

@attr.s(slots=True, frozen=True)
class KgtkLift(KgtkFormat):
    input_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))
    label_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))

    lift_column_names: typing.Optional[typing.List[str]] = \
        attr.ib(validator=attr.validators.optional(attr.validators.deep_iterable(member_validator=attr.validators.instance_of(str),
                                                                                 iterable_validator=attr.validators.instance_of(list))))
    output_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))
 
    node1_column_name: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)
    label_column_name: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)
    node2_column_name: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)

    label_column_value: str = attr.ib(validator=attr.validators.instance_of(str), default="label")
    lifted_column_suffix: str = attr.ib(validator=attr.validators.instance_of(str), default=";label")

    remove_label_records: bool = attr.ib(validator=attr.validators.instance_of(bool), default=True)
    suppress_duplicate_labels: bool = attr.ib(validator=attr.validators.instance_of(bool), default=True)
    sort_lifted_labels: bool = attr.ib(validator=attr.validators.instance_of(bool), default=True)
    suppress_empty_columns: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    ok_if_no_labels: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    input_is_presorted: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    labels_are_presorted: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    # TODO: add rewind logic here and KgtkReader

    # TODO: find working validators
    # value_options: typing.Optional[KgtkValueOptions] = attr.ib(attr.validators.optional(attr.validators.instance_of(KgtkValueOptions)), default=None)
    reader_options: typing.Optional[KgtkReaderOptions]= attr.ib(default=None)
    value_options: typing.Optional[KgtkValueOptions] = attr.ib(default=None)

    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    very_verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    def build_lift_column_idxs(self, kr: KgtkReader)->typing.List[int]:
        lift_column_idxs: typing.List[int] = [ ]
        if self.lift_column_names is not None and len(self.lift_column_names) > 0:
            # Process a custom list of columns to be lifted.
            lift_column_name: str
            for lift_column_name in self.lift_column_names:
                if lift_column_name not in kr.column_name_map:
                    raise ValueError("Unknown lift column %s." % lift_column_name)
                lift_column_idxs.append(kr.column_name_map[lift_column_name])
        else:
            # Use the edge file key columns if they exist.
            if kr.node1_column_idx >= 0:
                lift_column_idxs.append(kr.node1_column_idx)
            if kr.label_column_idx >= 0:
                lift_column_idxs.append(kr.label_column_idx)
            if kr.node2_column_idx >= 0:
                lift_column_idxs.append(kr.node2_column_idx)

        # Verify that we found some columns to lift:
        if len(lift_column_idxs) == 0:
            raise ValueError("No lift columns found.")

        return lift_column_idxs

    def lookup_node1_column_idx(self, kr: KgtkReader)->int:
        node1_column_idx: int
        if self.node1_column_name is None:
            if kr.node1_column_idx < 0:
                raise ValueError("No node1 column index.")
            node1_column_idx = kr.node1_column_idx
        else:
            if self.node1_column_name not in kr.column_name_map:
                raise ValueError("Node1 column `%s` not found." % self.node1_column_name)
            node1_column_idx = kr.column_name_map[self.node1_column_name]
        return node1_column_idx

    def lookup_label_column_idx(self, kr: KgtkReader)->int:
        label_column_idx: int
        if self.label_column_name is None:
            if kr.label_column_idx < 0:
                raise ValueError("No label column index.")
            label_column_idx = kr.label_column_idx
        else:
            if self.label_column_name not in kr.column_name_map:
                raise ValueError("Label column `%s` not found." % self.label_column_name)
            label_column_idx = kr.column_name_map[self.label_column_name]
        return label_column_idx

    def lookup_node2_column_idx(self, kr: KgtkReader)->int:
        node2_column_idx: int
        if self.node2_column_name is None:
            if kr.node2_column_idx < 0:
                raise ValueError("No node2 column index.")
            node2_column_idx = kr.node2_column_idx
        else:
            if self.node2_column_name not in kr.column_name_map:
                raise ValueError("Node2 column `%s` not found." % self.node2_column_name)
            node2_column_idx = kr.column_name_map[self.node2_column_name]

        return node2_column_idx

    def lookup_label_table_idxs(self, kr: KgtkReader)->typing.Tuple[int, int, int]:
        node1_column_idx: int = self.lookup_node1_column_idx(kr)
        label_column_idx: int = self.lookup_label_column_idx(kr)
        node2_column_idx: int = self.lookup_node2_column_idx(kr)

        return node1_column_idx, label_column_idx, node2_column_idx

    def load_labels(self,
                    kr: KgtkReader,
                    path: Path,
    )->typing.Tuple[typing.Mapping[str, str], typing.List[typing.List[str]]]:
        input_rows: typing.List[typing.List[str]] = [ ]
        labels: typing.MutableMapping[str, str] = { }

        node1_column_idx: int
        label_column_idx: int
        node2_column_idx: int
        node1_column_idx, label_column_idx, node2_column_idx = self.lookup_label_table_idxs(kr)

        if self.verbose:
            print("Loading labels from %s" % path, file=self.error_file, flush=True)
        key: str
        row: typing.List[str]
        for row in kr:
            if row[label_column_idx] == self.label_column_value:
                # This is a label definition row.
                label_key = row[node1_column_idx]
                label_value: str = row[node2_column_idx]
                if len(label_value) > 0:
                    if label_key in labels:
                        # This label already exists in the table.
                        if self.suppress_duplicate_labels:
                            # Build a list eliminating duplicate elements.
                            # print("Merge '%s' and '%s'" % (key_value, labels[key]), file=self.error_file, flush=True)
                            labels[label_key] = KgtkValue.merge_values(labels[label_key], label_value)
                        else:
                            labels[label_key] = KgtkFormat.LIST_SEPARATOR.join((labels[label_key], label_value))
                    else:
                        # This is the first instance of this label definition.
                        labels[label_key] = label_value
                if not self.remove_label_records:
                    input_rows.append(row)
            else:
                input_rows.append(row)
        return labels, input_rows
                
    def load_input_keeping_label_records(self,
                                         kr: KgtkReader,
                                         path: Path,
    )-> typing.List[typing.List[str]]:
        input_rows: typing.List[typing.List[str]] = [ ]

        if self.verbose:
            print("Loading input rows with labels from %s" % path, file=self.error_file, flush=True)
        row: typing.List[str]
        for row in kr:
            input_rows.append(row)
        return input_rows

    def load_input_removing_label_records(self,
                                          kr: KgtkReader,
                                          path: Path,
    )-> typing.List[typing.List[str]]:
        input_rows: typing.List[typing.List[str]] = [ ]

        if self.verbose:
            print("Loading input rows without labels from %s" % path, file=self.error_file, flush=True)
        row: typing.List[str]

        label_column_idx: int = self.lookup_label_column_idx(kr)
        for row in kr:
            if row[label_column_idx] != self.label_column_value:
                input_rows.append(row)

        return input_rows

    def load_input(self,
                   kr: KgtkReader,
                   path: Path,
    )-> typing.List[typing.List[str]]:
        if self.remove_label_records:
            return self.load_input_removing_label_records(kr, path)
        else:
            return self.load_input_keeping_label_records(kr, path)

    def build_lifted_column_idxs(self,
                                 kr: KgtkReader,
                                 lift_column_idxs: typing.List[int],
                                 input_rows: typing.List[typing.List[str]],
                                 labels: typing.Mapping[str, str],
                                 label_column_idx: int,
    )->typing.List[int]:
        """
        Build the lifted column indexes, suppressing those columns
        for which there are are no label values.  This requires a
        pass through the input data.
        """
        if self.verbose:
            print("Checking for empty columns", file=self.error_file, flush=True)
        lift_column_idxs_empties: typing.List[int] = lift_column_idxs.copy()
        lift_column_idx: int
        # Scan the input file, checking for empty output columns.
        for row in input_rows:
            if label_column_idx >= 0:
                if row[label_column_idx] == self.label_column_value:
                    # Skip label records if they have been saved.
                    continue
            idx: int
            restart: bool = True
            while restart:
                # The restart mechanism compensates for modifying
                # lift_column_idxs_empties inside the for loop, at the
                # expense of potentially double testing some items.
                restart = False
                for idx, lift_column_idx in enumerate(lift_column_idxs_empties):
                    item: str = row[lift_column_idx]
                    if item in labels:
                        lift_column_idxs_empties.pop(idx)
                        restart = True
                        break
            if len(lift_column_idxs_empties) == 0:
                break

        if self.verbose:
            if len(lift_column_idxs_empties) == 0:
                print("No lifted columns are empty", file=self.error_file, flush=True)
            else:
                lift_column_names_empties: typing.List[str] = [ ]
                for idx in lift_column_idxs_empties:
                    lift_column_names_empties.append(kr.column_names[idx])
                print("Unlifted columns: %s" % " ".join(lift_column_names_empties), file=self.error_file, flush=True)

        lifted_column_idxs: typing.List[int] = [ ]
        for lift_column_idx in lift_column_idxs:
            if lift_column_idx not in lift_column_idxs_empties:
                lifted_column_idxs.append(lift_column_idx)            
        return lifted_column_idxs

    def write_output_row(self,
                         ew: KgtkWriter,
                         row: typing.List[str],
                         new_columns: int,
                         label_column_idx: int,
                         labels: typing.Mapping[str, str],
                         lifted_column_idxs: typing.List[int],
                         lifted_output_column_idxs: typing.List[int]):
        output_row: typing.List[str] = row.copy()
        if new_columns > 0:
            output_row.extend([""] * new_columns)
                
        if label_column_idx >= 0 and row[label_column_idx] == self.label_column_value:
            # Don't lift label columns, if we have stored labels in the input records.
            pass
        else:
            # Lift the specified columns in this row.
            lifted_column_idx: int
            for idx, lifted_column_idx in enumerate(lifted_column_idxs):
                lifted_value: str = row[lifted_column_idx]
                if lifted_value in labels:
                    output_row[lifted_output_column_idxs[idx]] = labels[row[lifted_column_idx]]

        ew.write(output_row)
        return

    def build_output_column_names(self, ikr: KgtkReader, lifted_column_idxs: typing.List[int])->typing.Tuple[typing.List[str], typing.List[int]]:
        # Build the output column names.
        output_column_names: typing.List[str] = ikr.column_names.copy()
        lifted_output_column_idxs: typing.List[int] = [ ]
        for idx in lifted_column_idxs:
            lifted_column_name: str = ikr.column_names[idx] + self.lifted_column_suffix
            if lifted_column_name in ikr.column_name_map:
                # Overwrite an existing lifted output column.
                #
                # TODO: DO we want to control whether or not we overwrite existing columns?
                lifted_output_column_idxs.append(ikr.column_name_map[lifted_column_name])
            else:
                # Append a new lifted output column.
                lifted_output_column_idxs.append(len(output_column_names))
                output_column_names.append(lifted_column_name)
        return output_column_names, lifted_output_column_idxs

    def open_output_writer(self,
                           ikr: KgtkReader,
                           lifted_column_idxs: typing.List[int]
    )->typing.Tuple[KgtkWriter, typing.List[int]]:
        # Build the output column names.
        output_column_names: typing.List[str]
        lifted_output_column_idxs: typing.List[int]
        output_column_names, lifted_output_column_idxs = self.build_output_column_names(ikr, lifted_column_idxs)

        if self.verbose:
            print("Opening the output file: %s" % self.output_file_path, file=self.error_file, flush=True)
        ew: KgtkWriter = KgtkWriter.open(output_column_names,
                                         self.output_file_path,
                                         mode=KgtkWriter.Mode[ikr.mode.name],
                                         require_all_columns=False,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=True,
                                         gzip_in_parallel=False,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose)        

        return ew, lifted_output_column_idxs

    def process_in_memory(self, ikr: KgtkReader, lkr: typing.Optional[KgtkReader]):
        """
        Process the lift using in-memory buffering.  The labels will added to a
        dict in memory, and depending upon the options selected, the input
        rows may be kept on a list in memory.

        """
        if self.verbose:
            print("Lifting with in-memory buffering.", file=self.error_file, flush=True)
        lift_column_idxs: typing.List[int] = self.build_lift_column_idxs(ikr)

        labels: typing.Mapping[str, str] = { }
        input_rows: typing.Optional[typing.List[typing.List[str]]] = None
        label_column_idx: int = -1

        # Extract the labels, and maybe store the input rows.
        if lkr is not None and self.label_file_path is not None:
            # Read the label file.
            if self.verbose:
                print("Loading labels from the label file.", file=self.error_file, flush=True)
            # We don't need to worry about label records in the input file.
            labels, _ = self.load_labels(lkr, self.label_file_path)
        else:
            if self.verbose:
                print("Loading labels and reading data from the input file.", file=self.error_file, flush=True)
            # Read the input file, extracting the labels. The label
            # records may or may not be saved in the input rows, depending
            # upon whether we plan to pass them throuhg to the output.
            labels, input_rows = self.load_labels(ikr, self.input_file_path)
            # Save the label column index in the input file.  We will use
            # this if we pass store label records through to output.
            label_column_idx = self.lookup_label_column_idx(ikr)

        label_count: int = len(labels)
        if label_count == 0 and not self.ok_if_no_labels:
            raise ValueError("No labels were found.")

        lifted_column_idxs: typing.List[int]
        if self.suppress_empty_columns:
            if input_rows is None:
                # We need to read the input records now in order to determine
                # which lifted columns must be suppressed.
                if self.verbose:
                    print("Reading input data to suppress empty columns.", file=self.error_file, flush=True)
                input_rows = self.load_input(ikr, self.input_file_path)
            lifted_column_idxs = self.build_lifted_column_idxs(ikr, lift_column_idxs, input_rows, labels, label_column_idx)
        else:
            # Lift all the candidate columns.
            lifted_column_idxs = lift_column_idxs.copy()

        ew: KgtkWriter
        lifted_output_column_idxs: typing.List[int]
        ew, lifted_output_column_idxs = self.open_output_writer(ikr, lifted_column_idxs)

        if self.verbose:
            print("Writing output records", file=self.error_file, flush=True)

        new_columns: int = len(ew.column_names) - len(ikr.column_names)
        input_line_count: int = 0
        output_line_count: int = 0
        if input_rows is None:
            # We will make a single pass through the input file.
            if self.remove_label_records and label_column_idx >= 0:
                # Don't store the label records.
                for row in ikr:
                    if row[label_column_idx] != self.label_column_value:
                        input_line_count += 1
                        self.write_output_row(ew, row, new_columns, label_column_idx, labels, lifted_column_idxs, lifted_output_column_idxs)
                        output_line_count += 1
            else:
                # Store the label records.  Don't lift them, but write them in the output.
                for row in ikr:
                    input_line_count += 1
                    self.write_output_row(ew, row, new_columns, label_column_idx, labels, lifted_column_idxs, lifted_output_column_idxs)
                    output_line_count += 1
        else:
            # Use the stored input records.
            input_line_count = len(input_rows)
            for row in input_rows:
                self.write_output_row(ew, row, new_columns, label_column_idx, labels, lifted_column_idxs, lifted_output_column_idxs)
                output_line_count += 1

        if self.verbose:
            print("Read %d non-label input records." % (input_line_count), file=self.error_file, flush=True)
            print("%d labels were found." % (label_count), file=self.error_file, flush=True)
            print("Wrote %d records." % (output_line_count), file=self.error_file, flush=True)
        
        ew.close()
    
    def process_as_merge(self, ikr: KgtkReader, lkr: KgtkReader):
        """
        Process the lift as a merge between two sorted files.

        """
        if self.verbose:
            print("Merging sorted input and label files.", file=self.error_file, flush=True)
        lift_column_idxs: typing.List[int] = self.build_lift_column_idxs(ikr)
        if len(lift_column_idxs) != 1:
            raise ValueError("Expecting exactly one lift_column_idxs, got %d" % len(lift_column_idxs))

        ew: KgtkWriter
        lifted_output_column_idxs: typing.List[int]
        ew, lifted_output_column_idxs = self.open_output_writer(ikr, lift_column_idxs)

        new_columns: int = len(ew.column_names) - len(ikr.column_names)
        if new_columns not in (0, 1):
            raise ValueError("Expecing zero or one new columns, got %d" % new_columns)

        lift_column_idx: int = lift_column_idxs[0] # For convenience
        lifted_output_column_idx: int = lifted_output_column_idxs[0] # For convenience

        node1_column_idx: int
        label_column_idx: int
        node2_column_idx: int
        node1_column_idx, label_column_idx, node2_column_idx = self.lookup_label_table_idxs(lkr)
  
        current_label_row: typing.Optional[typing.List[str]] = None
        more_labels: bool = True
        # Read the first label record.
        try:
            current_label_row = lkr.nextrow()
        except StopIteration:
            more_labels = False

        input_line_count: int = 0

        # We carry last_value_to_lift and lifted_label_value over
        # iterations in case the input file has multiple records with
        # the same value to lift.
        last_value_to_lift: typing.Optional[str] = None
        lifted_label_value: str = ""

        if self.verbose:
            print("Processing the input records.", file=self.error_file, flush=True)

        row: typing.List[str]
        for row in ikr:
            input_line_count += 1
            value_to_lift: str = row[lift_column_idx]
            if last_value_to_lift is None or value_to_lift != last_value_to_lift:
                last_value_to_lift = value_to_lift
                lifted_label_value = ""

                # Read label records until we come to the first record that
                # has a node1 value equal to or greater than the value we we want to lift.
                while more_labels and current_label_row is not None and current_label_row[node1_column_idx] < value_to_lift:
                    try:
                        current_label_row = lkr.nextrow()
                    except StopIteration:
                        more_labels = False
                        break

                # While the label records have node1 values equal to the value we are trying to lift,
                # look for label values from the label file.
                while more_labels and current_label_row is not None and current_label_row[node1_column_idx] == value_to_lift:
                    if current_label_row[label_column_idx] == self.label_column_value:
                        label_value: str = current_label_row[node2_column_idx]
                        if len(label_value) > 0:
                            if len(lifted_label_value) > 0:
                                if self.suppress_duplicate_labels:
                                    lifted_label_value = KgtkValue.merge_values(lifted_label_value, label_value)
                                else:
                                    lifted_label_value = KgtkFormat.LIST_SEPARATOR.join((lifted_label_value, label_value))
                            else:
                                lifted_label_value = label_value

                    try:
                        current_label_row = lkr.nextrow()
                    except StopIteration:
                        more_labels = False
                        break

            output_row: typing.List[str] = row.copy()
            if new_columns > 0:
                output_row.append("")
            output_row[lifted_output_column_idx] = lifted_label_value
            ew.write(output_row)

        if more_labels:
            lkr.close()

        if self.verbose:
            print("Read %d input records." % (input_line_count), file=self.error_file, flush=True)
             
    
    def process(self):
        # Open the input file.
        if self.verbose:
            if self.input_file_path is not None:
                print("Opening the input file: %s" % self.input_file_path, file=self.error_file, flush=True)
            else:
                print("Reading the input data from stdin", file=self.error_file, flush=True)

        ikr: KgtkReader =  KgtkReader.open(self.input_file_path,
                                          error_file=self.error_file,
                                          options=self.reader_options,
                                          value_options = self.value_options,
                                          verbose=self.verbose,
                                          very_verbose=self.very_verbose,
        )

        # If supplied, open the label file.
        lkr: typing.Optional[KgtkReader] = None
        if self.label_file_path is not None:
            if self.verbose:
                if self.input_file_path is not None:
                    print("Opening the label file: %s" % self.label_file_path, file=self.error_file, flush=True)
                else:
                    print("Reading the label data from stdin", file=self.error_file, flush=True)

            lkr =  KgtkReader.open(self.label_file_path,
                                   error_file=self.error_file,
                                   options=self.reader_options,
                                   value_options = self.value_options,
                                   verbose=self.verbose,
                                   very_verbose=self.very_verbose,
            )

        if self.lift_column_names is not None and len(self.lift_column_names) == 1 and \
           not self.suppress_empty_columns and \
           self.input_is_presorted and \
           self.labels_are_presorted and \
           lkr is not None:
           self.process_as_merge(ikr, lkr)
        else:
            self.process_in_memory(ikr, lkr)

def main():
    """
    Test the KGTK lift processor.
    """
    parser: ArgumentParser = ArgumentParser()

    parser.add_argument(dest="input_file_path", help="The KGTK file with the input data", type=Path, default="-")

    parser.add_argument(      "--label-file", dest="label_file_path", help="A KGTK file with label records (default=%(default)s).", type=Path, default=None)

    parser.add_argument(      "--node1-name", dest="node1_column_name",
                              help="The name of the node1 column. (default=node1 or alias).", default=None)

    parser.add_argument(      "--label-name", dest="label_column_name",
                              help="The name of the label column. (default=label).", default=None)

    parser.add_argument(      "--node2-name", dest="node2_column_name",
                              help="The name of the node2 column. (default=node2 or alias).", default=None)

    parser.add_argument(      "--label-value", dest="label_column_value", help="The value in the label column. (default=%(default)s).", default="label")
    parser.add_argument(      "--lift-suffix", dest="lifted_column_suffix",
                              help="The suffix used for newly created columns. (default=%(default)s).", default=";label")

    parser.add_argument(      "--columns-to-lift", dest="lift_column_names", help="The columns to lift. (default=[node1, label, node2]).", nargs='*')

    parser.add_argument("-o", "--output-file", dest="output_file_path", help="The KGTK file to write (default=%(default)s).", type=Path, default="-")
    
    parser.add_argument(      "--remove-label-records", dest="remove_label_records",
                              help="If true, remove label records from the output. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--sort-lifted-labels", dest="sort_lifted_labels",
                              help="If true, sort lifted labels with lists. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--suppress-duplicate-labels", dest="suppress_duplicate_labels",
                              help="If true, suppress duplicate values in lifted labels with lists (implies sorting). (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--suppress-empty-columns", dest="suppress_empty_columns",
                              help="If true, do not create new columns that would be empty. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--ok-if-no-labels", dest="ok_if_no_labels",
                              help="If true, do not abort if no labels were found. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--input-file-is-presorted", dest="input_is_presorted",
                              help="If true, the input file is presorted on the column for which values are to be lifted. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--label-file-is-presorted", dest="labels_are_presorted",
                              help="If true, the label file is presorted on the node1 column. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)


    KgtkReader.add_debug_arguments(parser)
    # TODO: seperate reader options for the label file.
    KgtkReaderOptions.add_arguments(parser, mode_options=True)
    KgtkValueOptions.add_arguments(parser)

    args: Namespace = parser.parse_args()

    error_file: typing.TextIO = sys.stdout if args.errors_to_stdout else sys.stderr

    # Build the option structures.                                                                                                                          
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_args(args)
    value_options: KgtkValueOptions = KgtkValueOptions.from_args(args)

   # Show the final option structures for debugging and documentation.                                                                                             
    if args.show_options:
        print("input: %s" % str(args.input_file_path), file=error_file, flush=True)
        if args.label_file_path is not None:
            print("--label-file=%s" % str(args.label_file_path), file=error_file, flush=True)
        if args.node1_column_name is not None:
            print("--node1-name=%s" % args.node1_column_name, file=error_file, flush=True)
        if args.label_column_name is not None:
            print("--label-name=%s" % args.label_column_name, file=error_file, flush=True)
        if args.node2_column_name is not None:
            print("--node2-name=%s" % args.node2_column_name, file=error_file, flush=True)
        print("--label-value=%s" % args.label_column_value, file=error_file, flush=True)
        print("--lift-suffix=%s" % args.lifted_column_suffix, file=error_file, flush=True)
        if args.lift_column_names is not None and len(args.lift_column_names) > 0:
            print("--columns-to-lift %s" % " ".join(args.lift_column_names), file=error_file, flush=True)
        print("--output-file=%s" % str(args.output_file_path), file=error_file, flush=True)
        print("--remove-label-records=%s" % str(args.remove_label_records))
        print("--sort-lifted-labels-labels=%s" % str(args.sort_lifted_labels))
        print("--suppress-duplicate-labels=%s" % str(args.suppress_duplicate_labels))
        print("--suppress-empty-columns=%s" % str(args.suppress_empty_columns))
        print("--ok-if-no-labels=%s" % str(args.ok_if_no_labels))
        print("--input-file-is-presorted=%s" % str(args.input_is_presorted))
        print("--label-file-is-presorted=%s" % str(args.labels_are_presorted))
        reader_options.show(out=error_file)
        value_options.show(out=error_file)

    kl: KgtkLift = KgtkLift(
        input_file_path=args.input_file_path,
        label_file_path=args.label_file_path,
        node1_column_name=args.node1_column_name,
        label_column_name=args.label_column_name,
        node2_column_name=args.node2_column_name,
        label_column_value=args.label_column_value,
        lifted_column_suffix=args.lifted_column_suffix,
        lift_column_names=args.lift_column_names,
        output_file_path=args.output_file_path,
        remove_label_records=args.remove_label_records,
        sort_lifted_labels=args.sort_lifted_labels,
        suppress_duplicate_labels=args.suppress_duplicate_labels,
        suppress_empty_columns=args.suppress_empty_columns,
        ok_if_no_labels=args.ok_if_no_labels,
        input_is_presorted=args.input_is_presorted,
        labels_are_presorted=args.labels_are_presorted,
        reader_options=reader_options,
        value_options=value_options,
        error_file=error_file,
        verbose=args.verbose,
        very_verbose=args.very_verbose)

    kl.process()

if __name__ == "__main__":
    main()



"""Filter a KGTK file based on whether one or more records exist in a second
KGTK file with matching values for one or more fields.

TODO: Need KgtkWriterOptions
"""

from argparse import Namespace, SUPPRESS
from pathlib import Path
import sys
import typing

from kgtk.cli_argparse import KGTKArgumentParser
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.join.unique import Unique
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

def parser():
    return {
        'help': 'Count unique values in a column.',
        'description': 'Count the unique values in a column in a KGTK file. Write the unique values and counts as a new KGTK file.' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert unique --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_argument(      "input_kgtk_file", nargs="?", help="The KGTK file to filter. May be omitted or '-' for stdin.", type=Path)

    parser.add_argument(      "--column", dest="column_name",
                              help="The column to count unique values (required).", required=True)

    parser.add_argument(      "--empty", dest="empty_value", help="A value to substitute for empty values (default=%(default)s).", default="")

    parser.add_argument("-o", "--output-file", dest="output_kgtk_file", help="The KGTK file to write (required).", type=Path, default=None)

    parser.add_argument(      "--label", dest="label_value", help="The output file label column value (default=%(default)s).", default="count")

    # TODO: use an emum
    parser.add_argument(      "--format", dest="output_format", help=h("The output file format and mode (default=%(default)s)."),
                              default="edge", choices=["edge", "node"])

    parser.add_argument(      "--prefix", dest="prefix", help=h("The value prefix (default=%(default)s)."), default="")

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_kgtk_file: typing.Optional[Path],
        output_kgtk_file: typing.Optional[Path],

        column_name: str,
        empty_value: str = "",
        label_value: str = "count",

        output_format: str = "edge",
        prefix: str = "",

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from kgtk.exceptions import KGTKException


    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("input: %s" % (str(input_kgtk_file) if input_kgtk_file is not None else "-"), file=error_file)
        print("--output-file=%s" % (str(output_kgtk_file) if output_kgtk_file is not None else "-"), file=error_file)
        print("--column=%s" % str(column_name), file=error_file)
        print("--empty=%s" % str(empty_value), file=error_file)
        print("--label=%s" % str(label_value), file=error_file)
        print("--format=%s" % output_format, file=error_file)
        print("--prefix=%s" % prefix, file=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:
        uniq: Unique = Unique(
            input_file_path=input_kgtk_file,
            output_file_path=output_kgtk_file,
            column_name=column_name,
            label_value=label_value,
            empty_value=empty_value,
            output_format=output_format,
            prefix=prefix,
            reader_options=reader_options,
            value_options=value_options,
            error_file=error_file,
            verbose=verbose,
            very_verbose=very_verbose,
        )
        
        uniq.process()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




"""
Write a KGTK edge or node file in TSV format.

"""

from argparse import ArgumentParser
import attr
import bz2
from enum import Enum
import gzip
import json
import lz4 # type: ignore
import lzma
from pathlib import Path
from multiprocessing import Queue
import sys
import typing

from kgtk.kgtkformat import KgtkFormat
from kgtk.io.kgtkbase import KgtkBase
from kgtk.io.kgtkreader import KgtkReader
from kgtk.utils.enumnameaction import EnumNameAction
from kgtk.utils.gzipprocess import GzipProcess
from kgtk.utils.validationaction import ValidationAction

@attr.s(slots=True, frozen=False)
class KgtkWriter(KgtkBase):
    GZIP_QUEUE_SIZE_DEFAULT: int = GzipProcess.GZIP_QUEUE_SIZE_DEFAULT

    # TODO: use an enum
    OUTPUT_FORMAT_CSV: str = "csv"
    OUTPUT_FORMAT_JSON: str = "json"
    OUTPUT_FORMAT_JSON_MAP: str = "json-map"
    OUTPUT_FORMAT_JSON_MAP_COMPACT: str = "json-map-compact"
    OUTPUT_FORMAT_JSONL: str = "jsonl"
    OUTPUT_FORMAT_JSONL_MAP: str = "jsonl-map"
    OUTPUT_FORMAT_JSONL_MAP_COMPACT: str = "jsonl-map-compact"
    OUTPUT_FORMAT_KGTK: str = "kgtk"
    OUTPUT_FORMAT_MD: str = "md"

    OUTPUT_FORMAT_CHOICES: typing.List[str] = [
        OUTPUT_FORMAT_CSV,
        OUTPUT_FORMAT_JSON,
        OUTPUT_FORMAT_JSON_MAP,
        OUTPUT_FORMAT_JSON_MAP_COMPACT,
        OUTPUT_FORMAT_JSONL,
        OUTPUT_FORMAT_JSONL_MAP,
        OUTPUT_FORMAT_JSONL_MAP_COMPACT,
        OUTPUT_FORMAT_KGTK,
        OUTPUT_FORMAT_MD,
    ]
    OUTPUT_FORMAT_DEFAULT: str = OUTPUT_FORMAT_KGTK

    file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))
    file_out: typing.TextIO = attr.ib() # Todo: validate TextIO
    column_separator: str = attr.ib(validator=attr.validators.instance_of(str))
    column_names: typing.List[str] = attr.ib(validator=attr.validators.deep_iterable(member_validator=attr.validators.instance_of(str),
                                                                                     iterable_validator=attr.validators.instance_of(list)))
    column_name_map: typing.Mapping[str, int] = attr.ib(validator=attr.validators.deep_mapping(key_validator=attr.validators.instance_of(str),
                                                                                               value_validator=attr.validators.instance_of(int)))

    # Use these names in the output file, but continue to use
    # column_names for shuffle lists.
    output_column_names: typing.List[str] = \
        attr.ib(validator=attr.validators.deep_iterable(member_validator=attr.validators.instance_of(str),
                                                        iterable_validator=attr.validators.instance_of(list)))

    # For convenience, the count of columns. This is the same as len(column_names).
    column_count: int = attr.ib(validator=attr.validators.instance_of(int))

    # Require or fill trailing fields?
    require_all_columns: bool = attr.ib(validator=attr.validators.instance_of(bool))
    prohibit_extra_columns: bool = attr.ib(validator=attr.validators.instance_of(bool))
    fill_missing_columns: bool = attr.ib(validator=attr.validators.instance_of(bool))

    # How should header errors be processed?
    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    header_error_action: ValidationAction = attr.ib(validator=attr.validators.instance_of(ValidationAction), default=ValidationAction.EXIT)

    # Other implementation options?
    gzip_in_parallel: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    gzip_thread: typing.Optional[GzipProcess] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(GzipProcess)), default=None)
    gzip_queue_size: int = attr.ib(validator=attr.validators.instance_of(int), default=GZIP_QUEUE_SIZE_DEFAULT)

    output_format: str = attr.ib(validator=attr.validators.instance_of(str), default=OUTPUT_FORMAT_DEFAULT) # TODO: use an enum

    line_count: int = attr.ib(validator=attr.validators.instance_of(int), default=0)

    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    very_verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    class Mode(Enum):
        """
        There are four file writing modes:
        """
        NONE = 0 # Enforce neither edge nor node file required columns
        EDGE = 1 # Enforce edge file required columns
        NODE = 2 # Enforce node file required columns
        AUTO = 3 # Automatically decide whether to enforce edge or node file required columns

    @classmethod
    def open(cls,
             column_names: typing.List[str],
             file_path: typing.Optional[Path],
             who: str = "output",
             require_all_columns: bool = True,
             prohibit_extra_columns: bool = True,
             fill_missing_columns: bool = False,
             error_file: typing.TextIO = sys.stderr,
             header_error_action: ValidationAction = ValidationAction.EXIT,
             gzip_in_parallel: bool = False,
             gzip_queue_size: int = GZIP_QUEUE_SIZE_DEFAULT,
             column_separator: str = KgtkFormat.COLUMN_SEPARATOR,
             mode: Mode = Mode.AUTO,
             output_format: typing.Optional[str] = None,
             output_column_names: typing.Optional[typing.List[str]] = None,
             old_column_names: typing.Optional[typing.List[str]] = None,
             new_column_names: typing.Optional[typing.List[str]] = None,
             verbose: bool = False,
             very_verbose: bool = False)->"KgtkWriter":
        if file_path is None or str(file_path) == "-":
            if verbose:
                print("KgtkWriter: writing stdout", file=error_file, flush=True)

            return cls._setup(column_names=column_names,
                              file_path=None,
                              who=who,
                              file_out=sys.stdout,
                              require_all_columns=require_all_columns,
                              prohibit_extra_columns=prohibit_extra_columns,
                              fill_missing_columns=fill_missing_columns,
                              error_file=error_file,
                              header_error_action=header_error_action,
                              gzip_in_parallel=gzip_in_parallel,
                              gzip_queue_size=gzip_queue_size,
                              column_separator=column_separator,
                              mode=mode,
                              output_format=output_format,
                              output_column_names=output_column_names,
                              old_column_names=old_column_names,
                              new_column_names=new_column_names,
                              verbose=verbose,
                              very_verbose=very_verbose,
            )
        
        if verbose:
            print("File_path.suffix: %s" % file_path.suffix, file=error_file, flush=True)

        if file_path.suffix in [".gz", ".bz2", ".xz", ".lz4"]:
            # TODO: find a better way to coerce typing.IO[Any] to typing.TextIO
            gzip_file: typing.TextIO
            if file_path.suffix == ".gz":
                if verbose:
                    print("KgtkWriter: writing gzip %s" % str(file_path), file=error_file, flush=True)
                gzip_file = gzip.open(file_path, mode="wt") # type: ignore
            elif file_path.suffix == ".bz2":
                if verbose:
                    print("KgtkWriter: writing bz2 %s" % str(file_path), file=error_file, flush=True)
                gzip_file = bz2.open(file_path, mode="wt") # type: ignore
            elif file_path.suffix == ".xz":
                if verbose:
                    print("KgtkWriter: writing lzma %s" % str(file_path), file=error_file, flush=True)
                gzip_file = lzma.open(file_path, mode="wt") # type: ignore
            elif file_path.suffix ==".lz4":
                if verbose:
                    print("KgtkWriter: writing lz4 %s" % str(file_path), file=error_file, flush=True)
                gzip_file = lz4.frame.open(file_or_path, mode="wt") # type: ignore
            else:
                # TODO: throw a better exception.
                raise ValueError("Unexpected file_path.suffiz = '%s'" % file_path.suffix)

            return cls._setup(column_names=column_names,
                              file_path=file_path,
                              who=who,
                              file_out=gzip_file,
                              require_all_columns=require_all_columns,
                              prohibit_extra_columns=prohibit_extra_columns,
                              fill_missing_columns=fill_missing_columns,
                              error_file=error_file,
                              header_error_action=header_error_action,
                              gzip_in_parallel=gzip_in_parallel,
                              gzip_queue_size=gzip_queue_size,
                              column_separator=column_separator,
                              mode=mode,
                              output_format=output_format,
                              output_column_names=output_column_names,
                              old_column_names=old_column_names,
                              new_column_names=new_column_names,
                              verbose=verbose,
                              very_verbose=very_verbose,
            )
            
        else:
            if output_format is None:
                # TODO: optionally stack these on top of compression
                if file_path.suffix == ".md":
                    output_format = "md"
                elif file_path.suffix == ".csv":
                    output_format = "csv"
                elif file_path.suffix == ".json":
                    output_format = "json"
                elif file_path.suffix == ".jsonl":
                    output_format = "jsonl"
                else:
                    output_format = "kgtk"

            if verbose:
                print("KgtkWriter: writing file %s" % str(file_path), file=error_file, flush=True)
            return cls._setup(column_names=column_names,
                              file_path=file_path,
                              who=who,
                              file_out=open(file_path, "w"),
                              require_all_columns=require_all_columns,
                              prohibit_extra_columns=prohibit_extra_columns,
                              fill_missing_columns=fill_missing_columns,
                              error_file=error_file,
                              header_error_action=header_error_action,
                              gzip_in_parallel=gzip_in_parallel,
                              gzip_queue_size=gzip_queue_size,
                              column_separator=column_separator,
                              mode=mode,
                              output_format=output_format,
                              output_column_names=output_column_names,
                              old_column_names=old_column_names,
                              new_column_names=new_column_names,
                              verbose=verbose,
                              very_verbose=very_verbose,
)
    
    @classmethod
    def _setup(cls,
               column_names: typing.List[str],
               file_path: typing.Optional[Path],
               who: str,
               file_out: typing.TextIO,
               require_all_columns: bool,
               prohibit_extra_columns: bool,
               fill_missing_columns: bool,
               error_file: typing.TextIO,
               header_error_action: ValidationAction,
               gzip_in_parallel: bool,
               gzip_queue_size: int,
               column_separator: str,
               mode: Mode = Mode.AUTO,
               output_format: typing.Optional[str] = None,
               output_column_names: typing.Optional[typing.List[str]] = None,
               old_column_names: typing.Optional[typing.List[str]] = None,
               new_column_names: typing.Optional[typing.List[str]] = None,
               verbose: bool = False,
               very_verbose: bool = False,
    )->"KgtkWriter":

        if output_format is None:
            output_format = cls.OUTPUT_FORMAT_DEFAULT
            if verbose:
                print("Defaulting the output format to %s" % output_format, file=error_file, flush=True)

        if output_format == cls.OUTPUT_FORMAT_CSV:
            column_separator = "," # What a cheat!
                
        if output_column_names is None:
            output_column_names = column_names
        else:
            # Rename all output columns.
            if len(output_column_names) != len(column_names):
                raise ValueError("%s: %d column names but %d output column names" % (who, len(column_names), len(output_column_names)))

        if old_column_names is not None or new_column_names is not None:
            # Rename selected output columns:
            if old_column_names is None or new_column_names is None:
                raise ValueError("%s: old/new column name mismatch" % who)
            if len(old_column_names) != len(new_column_names):
                raise ValueError("%s: old/new column name length mismatch: %d != %d" % (who, len(old_column_names), len(new_column_names)))

            # Rename columns in place.  Start by copyin the output column name
            # list so the changes don't inadvertantly propogate.
            output_column_names = output_column_names.copy()
            column_name: str
            idx: int
            for idx, column_name in enumerate(old_column_names):
                if column_name not in output_column_names:
                    raise ValueError("%s: old column names %s not in the output column names." % (who, column_name))
                output_column_names[output_column_names.index(column_name)] = new_column_names[idx]
                

        # Build a map from column name to column index.  This is used for
        # self.writemap(...)  and self.build_shuffle_list(...)
        column_name_map: typing.Mapping[str, int] = cls.build_column_name_map(column_names,
                                                                              header_line=column_separator.join(column_names),
                                                                              who=who,
                                                                              error_action=header_error_action,
                                                                              error_file=error_file)

        # Build a header line for error feedback:
        header: str = column_separator.join(output_column_names)

        # Build a map from output column name to column index.
        output_column_name_map: typing.Mapping[str, int] = cls.build_column_name_map(output_column_names,
                                                                                     header_line=header,
                                                                                     who=who,
                                                                                     error_action=header_error_action,
                                                                                     error_file=error_file)

        # Should we automatically determine if this is an edge file or a node file?
        is_edge_file: bool = False
        is_node_file: bool = False
        if mode is KgtkWriter.Mode.AUTO:
            # If we have a node1 (or alias) column, then this must be an edge file. Otherwise, assume it is a node file.
            node1_idx: int = cls.get_column_idx(cls.NODE1_COLUMN_NAMES, output_column_name_map,
                                                header_line=header,
                                                who=who,
                                                error_action=header_error_action,
                                                error_file=error_file,
                                                is_optional=True)
            is_edge_file = node1_idx >= 0
            is_node_file = not is_edge_file
        elif mode is KgtkWriter.Mode.EDGE:
            is_edge_file = True
        elif mode is KgtkWriter.Mode.NODE:
            is_node_file = True
        elif mode is KgtkWriter.Mode.NONE:
            pass
        
        # Validate that we have the proper columns for an edge or node file,
        # ignoring the result.
        cls.get_special_columns(output_column_name_map,
                                header_line=header,
                                who=who,
                                error_action=header_error_action,
                                error_file=error_file,
                                is_edge_file=is_edge_file,
                                is_node_file=is_node_file)

        gzip_thread: typing.Optional[GzipProcess] = None
        if gzip_in_parallel:
            if verbose:
                print("Starting the gzip process.", file=error_file, flush=True)
            gzip_thread = GzipProcess(file_out, Queue(gzip_queue_size))
            gzip_thread.start()

        kw: KgtkWriter = cls(file_path=file_path,
                             file_out=file_out,
                             column_separator=column_separator,
                             column_names=column_names,
                             column_name_map=column_name_map,
                             column_count=len(column_names),
                             require_all_columns=require_all_columns,
                             prohibit_extra_columns=prohibit_extra_columns,
                             fill_missing_columns=fill_missing_columns,
                             error_file=error_file,
                             header_error_action=header_error_action,
                             gzip_in_parallel=gzip_in_parallel,
                             gzip_thread=gzip_thread,
                             gzip_queue_size=gzip_queue_size,
                             output_format=output_format,
                             output_column_names=output_column_names,
                             line_count=1,
                             verbose=verbose,
                             very_verbose=very_verbose,
        )
        kw.write_header()
        return kw


    def join_csv(self, values: typing.List[str])->str:
        line: str = ""
        value: str
        for value in values:
            if '"' in value or ',' in value:
                value = '"' + '""'.join(value.split('"')) + '"'
            if len(line) > 0:
                line += ","
            line += value
        return line

    def join_md(self, values: typing.List[str])->str:
        line: str = "|"
        value: str
        for value in values:
            value = "\\|".join(value.split("|"))
            line += " " + value + " |"
        return line

    def json_map(self, values: typing.List[str], compact: bool = False)->typing.Mapping[str, str]:
        result: typing.MutableMapping[str, str] = { }
        idx: int
        value: str
        for idx, value in enumerate(values):
            if len(value) > 0 or not compact:
                result[self.output_column_names[idx]] = value
        return result

    def write_header(self):
        header: str
        header2: typing.Optional[str] = None

        # Contemplate a last-second rename of the columns
        column_names: typing.List[str]
        if self.output_column_names is not None:
            column_names = self.output_column_names
        else:
            column_names = self.column_names

        if self.output_format == self.OUTPUT_FORMAT_JSON:
            self.writeline("[")
            header = json.dumps(column_names, indent=None, separators=(',', ':')) + ","
        elif self.output_format == self.OUTPUT_FORMAT_JSON_MAP:
            self.writeline("[")
            return
        elif self.output_format == self.OUTPUT_FORMAT_JSON_MAP_COMPACT:
            self.writeline("[")
            return
        elif self.output_format == self.OUTPUT_FORMAT_JSONL:
            header = json.dumps(column_names, indent=None, separators=(',', ':'))
        elif self.output_format == self.OUTPUT_FORMAT_JSONL_MAP:
            return
        elif self.output_format == self.OUTPUT_FORMAT_JSONL_MAP_COMPACT:
            return
        elif self.output_format == self.OUTPUT_FORMAT_MD:
            header = "|"
            header2 = "|"
            col: str
            for col in column_names:
                col = "\\|".join(col.split("|"))
                header += " " + col + " |"
                header2 += " -- |"
            
        elif self.output_format in [self.OUTPUT_FORMAT_KGTK, self.OUTPUT_FORMAT_CSV]:
            header = self.column_separator.join(column_names)
        else:
            raise ValueError("KgtkWriter: header: Unrecognized output format '%s'." % self.output_format)

        # Write the column names to the first line.
        if self.verbose:
            print("header: %s" % header, file=self.error_file, flush=True)
        self.writeline(header)
        if header2 is not None:
            self.writeline(header2)

    def writeline(self, line: str):
        if self.gzip_thread is not None:
            self.gzip_thread.write(line + "\n") # Todo: use system end-of-line sequence?
        else:
            self.file_out.write(line + "\n") # Todo: use system end-of-line sequence?

    # Write the next list of edge values as a list of strings.
    # TODO: Convert integers, coordinates, etc. from Python types
    def write(self, values: typing.List[str],
              shuffle_list: typing.Optional[typing.List[int]]= None):

        if shuffle_list is not None:
            if len(shuffle_list) != len(values):
                # TODO: throw a better exception
                raise ValueError("The shuffle list is %d long but the values are %d long" % (len(shuffle_list), len(values)))

            shuffled_values: typing.List[str] = [""] * self.column_count
            idx: int
            for idx in range(len(shuffle_list)):
                shuffle_idx: int = shuffle_list[idx]
                if shuffle_idx >= 0:
                    shuffled_values[shuffle_idx] = values[idx]
            values = shuffled_values

        # Optionally fill missing trailing columns with empty values:
        if self.fill_missing_columns and len(values) < self.column_count:
            while len(values) < self.column_count:
                values.append("")

        # Optionally validate that the line contained the right number of columns:
        #
        # When we report line numbers in error messages, line 1 is the first line after the header line.
        line: str
        if self.require_all_columns and len(values) < self.column_count:
            line = self.column_separator.join(values)
            raise ValueError("Required %d columns in input line %d, saw %d: '%s'" % (self.column_count, self.line_count, len(values), line))
        if self.prohibit_extra_columns and len(values) > self.column_count:
            line = self.column_separator.join(values)
            raise ValueError("Required %d columns in input line %d, saw %d (%d extra): '%s'" % (self.column_count, self.line_count, len(values),
                                                                                                len(values) - self.column_count, line))
        if self.output_format == self.OUTPUT_FORMAT_KGTK:
            self.writeline(self.column_separator.join(values))
        elif self.output_format == self.OUTPUT_FORMAT_CSV:
            self.writeline(self.join_csv(values))
        elif self.output_format == self.OUTPUT_FORMAT_MD:
            self.writeline(self.join_md(values))
        elif self.output_format == self.OUTPUT_FORMAT_JSON:
            self.writeline(json.dumps(values, indent=None, separators=(',', ':')) + ",")
        elif self.output_format == self.OUTPUT_FORMAT_JSON_MAP:
            self.writeline(json.dumps(self.json_map(values), indent=None, separators=(',', ':')) + ",")
        elif self.output_format == self.OUTPUT_FORMAT_JSON_MAP_COMPACT:
            self.writeline(json.dumps(self.json_map(values, compact=True), indent=None, separators=(',', ':')) + ",")
        elif self.output_format == self.OUTPUT_FORMAT_JSONL:
            self.writeline(json.dumps(values, indent=None, separators=(',', ':')))
        elif self.output_format == self.OUTPUT_FORMAT_JSONL_MAP:
            self.writeline(json.dumps(self.json_map(values), indent=None, separators=(',', ':')))
        elif self.output_format == self.OUTPUT_FORMAT_JSONL_MAP_COMPACT:
            self.writeline(json.dumps(self.json_map(values, compact=True), indent=None, separators=(',', ':')))
        else:
            raise ValueError("Unrecognized output format '%s'." % self.output_format)

        self.line_count += 1
        if self.very_verbose:
            sys.stdout.write(".")
            sys.stdout.flush()

    def flush(self):
        if self.gzip_thread is None:
            self.file_out.flush()

    def close(self):
        if self.output_format == "json":
            if self.verbose:
                print("Closing the JSON list.", file=self.error_file, flush=True)
            self.writeline("]")

        if self.gzip_thread is not None:
            self.gzip_thread.close()
        else:
            self.file_out.close()


    def writemap(self, value_map: typing.Mapping[str, str]):
        """
        Write a map of values to the output file.
        """
        column_name: str

        # Optionally check for unexpected column names:
        if self.prohibit_extra_columns:
            for column_name in value_map.keys():
                if column_name not in self.column_name_map:
                    raise ValueError("Unexpected column name %s at data record %d" % (column_name, self.line_count))

        values: typing.List[str] = [ ]
        for column_name in self.column_names:
            if column_name in value_map:
                values.append(value_map[column_name])
            elif self.require_all_columns:
                # TODO: throw a better exception.
                raise ValueError("Missing column %s at data record %d" % (column_name, self.line_count))
            else:
                values.append("")
                
        self.write(values)

    def build_shuffle_list(self,
                           other_column_names: typing.List[str],
                           fail_on_unknown_column: bool = False)->typing.List[int]:
        results: typing.List[int] = [ ]
        column_name: str
        for column_name in other_column_names:
            if column_name in self.column_name_map:
                results.append(self.column_name_map[column_name])
            elif fail_on_unknown_column:
                # TODO: throw a better exception
                raise ValueError("Unknown column name %s when building shuffle list" % column_name)
            else:
                results.append(-1) # Means skip this column.
        return results
    
def main():
    """
    Test the KGTK edge file writer.

    TODO: full reader options.

    TODO:  --show-options
    """
    parser = ArgumentParser()
    parser.add_argument(dest="input_kgtk_file", help="The KGTK file to read", type=Path, nargs="?")
    parser.add_argument(dest="output_kgtk_file", help="The KGTK file to write", type=Path, nargs="?")
    parser.add_argument(      "--header-error-action", dest="header_error_action",
                              help="The action to take when a header error is detected  Only ERROR or EXIT are supported.",
                              type=ValidationAction, action=EnumNameAction, default=ValidationAction.EXIT)
    parser.add_argument(      "--gzip-in-parallel", dest="gzip_in_parallel", help="Execute gzip in a subthread.", action='store_true')
    parser.add_argument(      "--input-mode", dest="input_mode",
                              help="Determine the input KGTK file mode.", type=KgtkReader.Mode, action=EnumNameAction, default=KgtkReader.Mode.AUTO)
    parser.add_argument(      "--output-mode", dest="output_mode",
                              help="Determine the output KGTK file mode.", type=KgtkWriter.Mode, action=EnumNameAction, default=KgtkWriter.Mode.AUTO)
    parser.add_argument(      "--output-format", dest="output_format", help="The file format (default=kgtk)", type=str)
    parser.add_argument(      "--output-columns", dest="output_column_names", help="Rename all output columns. (default=%(default)s)", type=str, nargs='+')
    parser.add_argument(      "--old-columns", dest="old_column_names", help="Rename seleted output columns: old names. (default=%(default)s)", type=str, nargs='+')
    parser.add_argument(      "--new-columns", dest="new_column_names", help="Rename seleted output columns: new names. (default=%(default)s)", type=str, nargs='+')
    parser.add_argument("-v", "--verbose", dest="verbose", help="Print additional progress messages.", action='store_true')
    parser.add_argument(      "--very-verbose", dest="very_verbose", help="Print additional progress messages.", action='store_true')
    args = parser.parse_args()

    error_file: typing.TextIO = sys.stdout if args.errors_to_stdout else sys.stderr

    kr: KgtkReader = KgtkReader.open(args.input_kgtk_file,
                                     error_file=error_file,
                                     header_error_action=args.header_error_action,
                                     gzip_in_parallel=args.gzip_in_parallel,
                                     mode=args.input_mode,
                                     verbose=args.verbose, very_verbose=args.very_verbose)

    kw: KgtkWriter = KgtkWriter.open(kr.column_names,
                                     args.output_kgtk_file,
                                     error_file=error_file,
                                     gzip_in_parallel=args.gzip_in_parallel,
                                     header_error_action=args.header_error_action,
                                     mode=args.output_mode,
                                     output_format=args.output_format,
                                     output_column_names=args.output_column_names,
                                     old_column_names=args.old_column_names,
                                     new_column_names=args.new_column_names,
                                     verbose=args.verbose, very_verbose=args.very_verbose)

    line_count: int = 0
    row: typing.List[str]
    for row in kr:
        kw.write(row)
        line_count += 1
    kw.close()
    if args.verbose:
        print("Copied %d lines" % line_count, file=error_file, flush=True)


if __name__ == "__main__":
    main()
